"""
This module provides functionality for calculating influences for large amount of data.
The computation is based on a chunk computation model in the form of an instance of
[InfluenceFunctionModel][pydvl.influence.base_influence_model.InfluenceFunctionModel],
which is mapped over collection of chunks.
"""

import logging
from functools import partial
from typing import Generator, Iterable, Optional, Tuple, Type, Union

import distributed
from dask import array as da
from dask import delayed
from distributed import Client
from numpy.typing import NDArray

from .array import LazyChunkSequence, NestedLazyChunkSequence, NumpyConverter
from .base_influence_function_model import (
    InfluenceFunctionModel,
    InfluenceMode,
    TensorType,
    UnsupportedInfluenceModeException,
)

__all__ = [
    "DaskInfluenceCalculator",
    "SequentialInfluenceCalculator",
    "DisableClientSingleThreadCheck",
]

logger = logging.getLogger(__name__)


class InvalidDimensionChunksError(ValueError):
    def __init__(self, chunks: Tuple[Tuple[int, ...], ...]):
        msg = (
            f"Array must be un-chunked in every dimension but the first, got {chunks=}"
        )
        super().__init__(msg)


class UnalignedChunksError(ValueError):
    def __init__(self, chunk_sizes_x: Tuple[int, ...], chunk_sizes_y: Tuple[int, ...]):
        msg = (
            f"Arrays x and y must have the same chunking in the first dimension, got "
            f"{chunk_sizes_x=} and {chunk_sizes_y=}"
        )
        super().__init__(msg)


class ThreadSafetyViolationError(ValueError):
    def __init__(self):
        msg = (
            "The operation cannot proceed because the "
            "influence_function_model is not thread-safe, yet it is being used in a "
            "context that requires thread safety. Please ensure that all components "
            "are compatible with multi-threaded environments or use single-threaded "
            "workers. \n"
            "client = Client(threads_per_worker=1)"
        )
        super().__init__(msg)


class DisableClientSingleThreadCheck:
    """
    This type can be provided to the initialization of a DaskInfluenceCalculator
    instead of a distributed client object. It is useful in those scenarios, where
    the user want to disable the checking for thread-safety in the initialization
    phase, e.g. when using the single machine synchronous scheduler for debugging
    purposes.

    Example:
        ```python
        from pydvl.influence import DisableClientThreadingCheck

        da_calc = DaskInfluenceCalculator(if_model,
                                          TorchNumpyConverter(),
                                          DisableClientThreadingCheck)
        da_influences = da_calc.influences(da_x_test, da_y_test, da_x, da_y)
        da_influences.compute(scheduler='synchronous')
        ```
    """

    @staticmethod
    def warning_msg() -> str:
        msg = (
            "You have disabled client threading safety check. "
            "This is only advised for debugging "
            "purposes with the single machine synchronous scheduler. "
            "Disabling these checks can lead to unpredictable behavior and "
            "is only advised when you are fully aware of the implications."
        )
        return msg


class DaskInfluenceCalculator:
    """
    This class is designed to compute influences over
    [dask.array.Array][dask.array.Array] collections, leveraging the capabilities of
    Dask for distributed computing and parallel processing. It requires an influence
    computation model of type
    [InfluenceFunctionModel]
    [pydvl.influence.base_influence_model.InfluenceFunctionModel],
    which defines how influences are computed on a chunk of data.
    Essentially, this class functions by mapping the influence
    function model across the various chunks of a [dask.array.Array][dask.array.Array]
    collection.

    Args:
        influence_function_model: instance of type
            [InfluenceFunctionModel]
            [pydvl.influence.base_influence_model.InfluenceFunctionModel], that
            specifies the computation logic for influence on data chunks. It's a pivotal
            part of the calculator, determining how influence is computed and applied
            across the data array.
        converter: A utility for converting numpy arrays to TensorType objects,
            facilitating the interaction between numpy arrays and the influence
            function model.
        client: This parameter accepts either of two types:

            1. A distributed [Client][distributed.Client] object

            2. The special type [DisableClientSingleThreadCheck]
            [pydvl.influence.influence_calculator.DisableClientSingleThreadCheck], which
            serves as a flag to bypass certain checks.

            During initialization, the system verifies if all workers are operating in
            single-threaded mode when the provided influence_function_model is
            designated as not thread-safe (indicated by the `is_thread_safe` property).
            If this condition is not met, the initialization will raise a specific
            error, signaling a potential thread-safety conflict.

            To intentionally skip this safety check
            (e.g., for debugging purposes using the single machine synchronous
            scheduler), you can supply the [DisableClientSingleThreadCheck]
            [pydvl.influence.influence_calculator.DisableClientSingleThreadCheck] type.

    !!! Warning
        Make sure to set `threads_per_worker=1`, when using the distributed scheduler
        for computing, if your implementation of
        [InfluenceFunctionModel]
        [pydvl.influence.base_influence_model.InfluenceFunctionModel]
        is not thread-safe.
        ```python
        client = Client(threads_per_worker=1)
        ```
        For details on dask schedulers see the
        [official documentation](https://docs.dask.org/en/stable/scheduling.html).

    Example:
        ```python
        import torch
        from torch.utils.data import Dataset, DataLoader
        from pydvl.influence import DaskInfluenceCalculator
        from pydvl.influence.torch import CgInfluence
        from pydvl.influence.torch.util import (
        torch_dataset_to_dask_array,
        TorchNumpyConverter,
        )
        from distributed import Client

        # Possible some out of memory large Dataset
        train_data_set: Dataset = LargeDataSet(...)
        test_data_set: Dataset = LargeDataSet(...)

        train_dataloader = DataLoader(train_data_set)
        infl_model = CgInfluence(model, loss, hessian_regularization=0.01)
        infl_model = if_model.fit(train_dataloader)

        # wrap your input data into dask arrays
        chunk_size = 10
        da_x, da_y = torch_dataset_to_dask_array(train_data_set, chunk_size=chunk_size)
        da_x_test, da_y_test = torch_dataset_to_dask_array(test_data_set,
                                                           chunk_size=chunk_size)

        # use only one thread for scheduling, due to non-thread safety of some torch
        # operations
        client = Client(n_workers=4, threads_per_worker=1)

        infl_calc = DaskInfluenceCalculator(infl_model,
                                            TorchNumpyConverter(device=torch.device("cpu")),
                                            client)
        da_influences = infl_calc.influences(da_x_test, da_y_test, da_x, da_y)
        # da_influences is a dask.array.Array

        # trigger computation and write chunks to disk in parallel
        da_influences.to_zarr("path/or/url")
        ```

    """

    def __init__(
        self,
        influence_function_model: InfluenceFunctionModel,
        converter: NumpyConverter,
        client: Union[Client, Type[DisableClientSingleThreadCheck]],
    ):
        self._n_parameters = influence_function_model.n_parameters
        self.influence_function_model = influence_function_model
        self.numpy_converter = converter

        if isinstance(client, type(DisableClientSingleThreadCheck)):
            logger.warning(DisableClientSingleThreadCheck.warning_msg())
            self.influence_function_model = delayed(influence_function_model)
        elif isinstance(client, Client):
            self._validate_client(client, influence_function_model)
            self.influence_function_model = client.scatter(
                influence_function_model, broadcast=True
            )
        else:
            raise ValueError(
                "The 'client' parameter "
                "must either be a distributed.Client object or the"
                "type 'DisableClientSingleThreadCheck'."
            )

    @staticmethod
    def _validate_client(
        client: Client, influence_function_model: InfluenceFunctionModel
    ):
        info = client.scheduler_info()
        workers = info["workers"]
        all_single_threaded = all(
            worker["nthreads"] == 1 for worker in workers.values()
        )
        if not (influence_function_model.is_thread_safe or all_single_threaded):
            raise ThreadSafetyViolationError()

    @property
    def n_parameters(self):
        """
        Number of trainable parameters of the underlying model used in the batch
        computation
        """
        return self._n_parameters

    @staticmethod
    def _validate_dimensions_not_chunked(x: da.Array):
        """
        Check if all but the first dimension are not chunked
        """
        if any([len(c) > 1 for c in x.chunks[1:]]):
            raise InvalidDimensionChunksError(x.chunks)

    @staticmethod
    def _validate_aligned_chunking(x: da.Array, y: da.Array):
        """
        Check that the chunking in the first dimensions of the two input arrays
        are aligned
        """
        if x.chunks[0] != y.chunks[0]:
            raise UnalignedChunksError(x.chunks[0], y.chunks[0])

    def influence_factors(self, x: da.Array, y: da.Array) -> da.Array:
        r"""
        Computes the expression

        \[ H^{-1}\nabla_{\theta} \ell(y, f_{\theta}(x)) \]

        where the gradients are computed for the chunks of $(x, y)$.

        Args:
            x: model input to use in the gradient computations
            y: label tensor to compute gradients

        Returns:
            [dask.array.Array][dask.array.Array] representing the element-wise inverse
                Hessian matrix vector products for the provided batch.

        """

        self._validate_aligned_chunking(x, y)
        self._validate_dimensions_not_chunked(x)
        self._validate_dimensions_not_chunked(y)

        def func(x_numpy: NDArray, y_numpy: NDArray, model: InfluenceFunctionModel):
            factors = model.influence_factors(
                self.numpy_converter.from_numpy(x_numpy),
                self.numpy_converter.from_numpy(y_numpy),
            )
            return self.numpy_converter.to_numpy(factors)

        chunks = []
        for x_chunk, y_chunk, chunk_size in zip(
            x.to_delayed(), y.to_delayed(), x.chunks[0]
        ):
            chunk_shape = (chunk_size, self.n_parameters)
            chunk_array = da.from_delayed(
                delayed(func)(
                    x_chunk.squeeze().tolist(),
                    y_chunk.squeeze().tolist(),
                    self.influence_function_model,
                ),
                dtype=x.dtype,
                shape=chunk_shape,
            )
            chunks.append(chunk_array)

        return da.concatenate(chunks)

    def influences(
        self,
        x_test: da.Array,
        y_test: da.Array,
        x: Optional[da.Array] = None,
        y: Optional[da.Array] = None,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> da.Array:
        r"""
        Compute approximation of

        \[ \langle H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
        f_{\theta}(x_{\text{test}})), \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the case of up-weighting influence, resp.

        \[ \langle H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
        f_{\theta}(x_{\text{test}})),
        \nabla_{x} \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the perturbation type influence case. The computation is done block-wise
        for the chunks of the provided dask arrays.

        Args:
            x_test: model input to use in the gradient computations of
                $H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
                    f_{\theta}(x_{\text{test}}))$
            y_test: label tensor to compute gradients
            x: optional model input to use in the gradient computations
                $\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                resp. $\nabla_{x}\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                if None, use $x=x_{\text{test}}$
            y: optional label tensor to compute gradients
            mode: enum value of [InfluenceType]
                [pydvl.influence.base_influence_model.InfluenceType]

        Returns:
            [dask.array.Array][dask.array.Array] representing the element-wise scalar
                products for the provided batch.

        """

        self._validate_aligned_chunking(x_test, y_test)
        self._validate_dimensions_not_chunked(x_test)
        self._validate_dimensions_not_chunked(y_test)

        if (x is None) != (y is None):
            if x is None:
                raise ValueError(
                    "Providing labels y without providing model input x "
                    "is not supported"
                )
            if y is None:
                raise ValueError(
                    "Providing model input x without labels y is not supported"
                )
        elif x is not None:
            self._validate_aligned_chunking(x, y)
            self._validate_dimensions_not_chunked(x)
            self._validate_dimensions_not_chunked(y)
        else:
            x, y = x_test, y_test

        def func(
            x_test_numpy: NDArray,
            y_test_numpy: NDArray,
            x_numpy: NDArray,
            y_numpy: NDArray,
            model: InfluenceFunctionModel,
        ):
            values = model.influences(
                self.numpy_converter.from_numpy(x_test_numpy),
                self.numpy_converter.from_numpy(y_test_numpy),
                self.numpy_converter.from_numpy(x_numpy),
                self.numpy_converter.from_numpy(y_numpy),
                mode,
            )
            return self.numpy_converter.to_numpy(values)

        un_chunked_x_shapes = [s[0] for s in x_test.chunks[1:]]
        x_test_chunk_sizes = x_test.chunks[0]
        x_chunk_sizes = x.chunks[0]
        blocks = []
        block_shape: Tuple[int, ...]

        for x_test_chunk, y_test_chunk, test_chunk_size in zip(
            x_test.to_delayed(), y_test.to_delayed(), x_test_chunk_sizes
        ):
            row = []
            for x_chunk, y_chunk, chunk_size in zip(
                x.to_delayed(), y.to_delayed(), x_chunk_sizes  # type:ignore
            ):
                if mode == InfluenceMode.Up:
                    block_shape = (test_chunk_size, chunk_size)
                elif mode == InfluenceMode.Perturbation:
                    block_shape = (test_chunk_size, chunk_size, *un_chunked_x_shapes)
                else:
                    raise UnsupportedInfluenceModeException(mode)

                block_array = da.from_delayed(
                    delayed(func)(
                        x_test_chunk.squeeze().tolist(),
                        y_test_chunk.squeeze().tolist(),
                        x_chunk.squeeze().tolist(),
                        y_chunk.squeeze().tolist(),
                        self.influence_function_model,
                    ),
                    shape=block_shape,
                    dtype=x_test.dtype,
                )

                if mode == InfluenceMode.Perturbation:
                    n_dims = block_array.ndim
                    new_order = tuple(range(2, n_dims)) + (0, 1)
                    block_array = block_array.transpose(new_order)

                row.append(block_array)
            blocks.append(row)

        values_array = da.block(blocks)

        if mode == InfluenceMode.Perturbation:
            n_dims = values_array.ndim
            new_order = (n_dims - 2, n_dims - 1) + tuple(range(n_dims - 2))
            values_array = values_array.transpose(new_order)

        return values_array

    def influences_from_factors(
        self,
        z_test_factors: da.Array,
        x: da.Array,
        y: da.Array,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> da.Array:
        r"""
        Computation of

        \[ \langle z_{\text{test_factors}},
            \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the case of up-weighting influence, resp.

        \[ \langle z_{\text{test_factors}},
            \nabla_{x} \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the perturbation type influence case. The gradient is meant
        to be per sample of the batch $(x, y)$.

        Args:
            z_test_factors: pre-computed array, approximating
                $H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
                    f_{\theta}(x_{\text{test}}))$
            x: optional model input to use in the gradient computations
                $\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                resp. $\nabla_{x}\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                if None, use $x=x_{\text{test}}$
            y: optional label tensor to compute gradients
            mode: enum value of [InfluenceType]
                [pydvl.influence.twice_differentiable.InfluenceType]

        Returns:
          [dask.array.Array][dask.array.Array] representing the element-wise scalar
            product of the provided batch

        """
        self._validate_aligned_chunking(x, y)
        self._validate_dimensions_not_chunked(x)
        self._validate_dimensions_not_chunked(y)
        self._validate_dimensions_not_chunked(z_test_factors)

        def func(
            z_test_numpy: NDArray,
            x_numpy: NDArray,
            y_numpy: NDArray,
            model: InfluenceFunctionModel,
        ):
            ups = model.influences_from_factors(
                self.numpy_converter.from_numpy(z_test_numpy),
                self.numpy_converter.from_numpy(x_numpy),
                self.numpy_converter.from_numpy(y_numpy),
                mode=mode,
            )
            return self.numpy_converter.to_numpy(ups)

        un_chunked_x_shape = [s[0] for s in x.chunks[1:]]
        x_chunk_sizes = x.chunks[0]
        z_test_chunk_sizes = z_test_factors.chunks[0]
        blocks = []
        block_shape: Tuple[int, ...]

        for z_test_chunk, z_test_chunk_size in zip(
            z_test_factors.to_delayed(), z_test_chunk_sizes
        ):
            row = []
            for x_chunk, y_chunk, chunk_size in zip(
                x.to_delayed(), y.to_delayed(), x_chunk_sizes
            ):
                if mode == InfluenceMode.Perturbation:
                    block_shape = (z_test_chunk_size, chunk_size, *un_chunked_x_shape)
                elif mode == InfluenceMode.Up:
                    block_shape = (z_test_chunk_size, chunk_size)
                else:
                    raise UnsupportedInfluenceModeException(mode)

                block_array = da.from_delayed(
                    delayed(func)(
                        z_test_chunk.squeeze().tolist(),
                        x_chunk.squeeze().tolist(),
                        y_chunk.squeeze().tolist(),
                        self.influence_function_model,
                    ),
                    shape=block_shape,
                    dtype=z_test_factors.dtype,
                )

                if mode == InfluenceMode.Perturbation:
                    n_dims = block_array.ndim
                    new_order = tuple(range(2, n_dims)) + (0, 1)
                    block_array = block_array.transpose(*new_order)

                row.append(block_array)
            blocks.append(row)

        values_array = da.block(blocks)

        if mode == InfluenceMode.Perturbation:
            n_dims = values_array.ndim
            new_order = (n_dims - 2, n_dims - 1) + tuple(range(n_dims - 2))
            values_array = values_array.transpose(*new_order)

        return values_array

    @staticmethod
    def _get_client() -> Optional[distributed.Client]:
        try:
            return distributed.get_client()
        except ValueError:
            return None


class SequentialInfluenceCalculator:
    """
    This class serves as a simple wrapper for processing batches of data
    in a sequential manner. It is particularly useful in scenarios where parallel
    or distributed processing is not required or not feasible.
    The core functionality of this class is to apply a specified influence computation
    model, of type [InfluenceFunctionModel]
    [pydvl.influence.base_influence_model.InfluenceFunctionModel], to batches of data
    one at a time.

    Args:
        influence_function_model: An instance of type
                [InfluenceFunctionModel]
                [pydvl.influence.base_influence_model.InfluenceFunctionModel], that
                specifies the computation logic for influence on data chunks.

    Example:
        ```python
        from pydvl.influence import SequentialInfluenceCalculator
        from pydvl.influence.torch.util import (
        NestedTorchCatAggregator,
        TorchNumpyConverter,
        )
        from pydvl.influence.torch import CgInfluence

        batch_size = 10
        train_dataloader = DataLoader(..., batch_size=batch_size)
        test_dataloader = DataLoader(..., batch_size=batch_size)

        infl_model = CgInfluence(model, loss, hessian_regularization=0.01)
        infl_model = infl_model.fit(train_dataloader)

        infl_calc = SequentialInfluenceCalculator(if_model)

        # this does not trigger the computation
        lazy_influences = infl_calc.influences(test_dataloader, train_dataloader)

        # trigger computation and pull the result into main memory, result is the full
        # tensor for all combinations of the two loaders
        influences = lazy_influences.compute(aggregator=NestedTorchCatAggregator())
        # or
        # trigger computation and write results chunk-wise to disk using zarr in a
        # sequential manner
        lazy_influences.to_zarr("local_path/or/url", TorchNumpyConverter())
        ```
    """

    def __init__(
        self,
        influence_function_model: InfluenceFunctionModel,
    ):
        self.influence_function_model = influence_function_model

    def _influence_factors_gen(
        self, data_iterable: Iterable[Tuple[TensorType, TensorType]]
    ) -> Generator[TensorType, None, None]:
        for x, y in iter(data_iterable):
            yield self.influence_function_model.influence_factors(x, y)

    def influence_factors(
        self,
        data_iterable: Iterable[Tuple[TensorType, TensorType]],
    ) -> LazyChunkSequence:
        r"""
        Compute the expression

        \[ H^{-1}\nabla_{\theta} \ell(y, f_{\theta}(x)) \]

        where the gradient are computed for the chunks $(x, y)$ of the data_iterable in
        a sequential manner.

        Args:
            data_iterable: An iterable that returns tuples of tensors.
                Each tuple consists of a pair of tensors (x, y), representing input data
                and corresponding targets.

        Returns:
            A lazy data structure representing the chunks of the resulting tensor
        """
        tensors_gen_factory = partial(self._influence_factors_gen, data_iterable)
        return LazyChunkSequence(tensors_gen_factory)

    def _influences_gen(
        self,
        test_data_iterable: Iterable[Tuple[TensorType, TensorType]],
        train_data_iterable: Iterable[Tuple[TensorType, TensorType]],
        mode: InfluenceMode,
    ) -> Generator[Generator[TensorType, None, None], None, None]:

        for x_test, y_test in iter(test_data_iterable):
            yield (
                self.influence_function_model.influences(x_test, y_test, x, y, mode)
                for x, y in iter(train_data_iterable)
            )

    def influences(
        self,
        test_data_iterable: Iterable[Tuple[TensorType, TensorType]],
        train_data_iterable: Iterable[Tuple[TensorType, TensorType]],
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> NestedLazyChunkSequence:
        r"""
        Compute approximation of

        \[ \langle H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
        f_{\theta}(x_{\text{test}})), \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the case of up-weighting influence, resp.

        \[ \langle H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
        f_{\theta}(x_{\text{test}})),
        \nabla_{x} \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the perturbation type influence case. The computation is done block-wise for
        the chunks of the provided
        data iterables and aggregated into a single tensor in memory.

        Args:
            test_data_iterable: An iterable that returns tuples of tensors.
                Each tuple consists of a pair of tensors (x, y), representing input data
                and corresponding targets.
            train_data_iterable: An iterable that returns tuples of tensors.
                Each tuple consists of a pair of tensors (x, y), representing input data
                and corresponding targets.
            mode: enum value of [InfluenceType]
                [pydvl.influence.base_influence_model.InfluenceType]

        Returns:
            A lazy data structure representing the chunks of the resulting tensor

        """
        nested_tensor_gen_factory = partial(
            self._influences_gen,
            test_data_iterable,
            train_data_iterable,
            mode,
        )

        return NestedLazyChunkSequence(nested_tensor_gen_factory)

    def _influences_from_factors_gen(
        self,
        z_test_factors: Iterable[TensorType],
        train_data_iterable: Iterable[Tuple[TensorType, TensorType]],
        mode: InfluenceMode,
    ):

        for z_test_factor in iter(z_test_factors):
            if isinstance(z_test_factor, list) or isinstance(z_test_factor, tuple):
                z_test_factor = z_test_factor[0]
            yield (
                self.influence_function_model.influences_from_factors(
                    z_test_factor, x, y, mode
                )
                for x, y in iter(train_data_iterable)
            )

    def influences_from_factors(
        self,
        z_test_factors: Iterable[TensorType],
        train_data_iterable: Iterable[Tuple[TensorType, TensorType]],
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> NestedLazyChunkSequence:
        r"""
        Computation of

        \[ \langle z_{\text{test_factors}}, \nabla_{\theta} \ell(y, f_{\theta}(x))
            \rangle \]

        for the case of up-weighting influence, resp.

        \[ \langle z_{\text{test_factors}}, \nabla_{x} \nabla_{\theta}
            \ell(y, f_{\theta}(x)) \rangle \]

        for the perturbation type influence case. The gradient is meant to be per sample
        of the batch $(x, y)$.

        Args:
            z_test_factors: Pre-computed iterable of tensors, approximating
                $H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
                    f_{\theta}(x_{\text{test}}))$
            train_data_iterable: An iterable that returns tuples of tensors.
                Each tuple consists of a pair of tensors (x, y), representing input data
                and corresponding targets.
            mode: enum value of [InfluenceType]
                [pydvl.influence.twice_differentiable.InfluenceType]

        Returns:
          A lazy data structure representing the chunks of the resulting tensor

        """
        nested_tensor_gen = partial(
            self._influences_from_factors_gen,
            z_test_factors,
            train_data_iterable,
            mode,
        )
        return NestedLazyChunkSequence(nested_tensor_gen)



from typing import Dict, Iterable, Union

import numpy as np
from numpy.typing import NDArray
from tqdm.auto import tqdm

from pydvl.utils import Utility
from pydvl.value.result import ValuationResult

__all__ = ["compute_removal_score"]


def compute_removal_score(
    u: Utility,
    values: ValuationResult,
    percentages: Union[NDArray[np.float_], Iterable[float]],
    *,
    remove_best: bool = False,
    progress: bool = False,
) -> Dict[float, float]:
    r"""Fits model and computes score on the test set after incrementally removing
    a percentage of data points from the training set, based on their values.

    Args:
        u: Utility object with model, data, and scoring function.
        values: Data values of data instances in the training set.
        percentages: Sequence of removal percentages.
        remove_best: If True, removes data points in order of decreasing valuation.
        progress: If True, display a progress bar.

    Returns:
        Dictionary that maps the percentages to their respective scores.
    """
    # Sanity checks
    if np.any([x >= 1.0 or x < 0.0 for x in percentages]):
        raise ValueError("All percentages should be in the range [0.0, 1.0)")

    if len(values) != len(u.data.indices):
        raise ValueError(
            f"The number of values, {len(values) }, should be equal to the number of data indices, {len(u.data.indices)}"
        )

    scores = {}

    # We sort in descending order if we want to remove the best values
    values.sort(reverse=remove_best)

    for pct in tqdm(percentages, disable=not progress, desc="Removal Scores"):
        n_removal = int(pct * len(u.data))
        indices = values.indices[n_removal:]
        score = u(indices)
        scores[pct] = score
    return scores



import os
from typing import Any, Dict, Optional

from pydvl.utils.caching.base import CacheBackend

__all__ = ["InMemoryCacheBackend"]


class InMemoryCacheBackend(CacheBackend):
    """In-memory cache backend that stores results in a dictionary.

    Implements the CacheBackend interface for an in-memory-based cache.
    Stores cache entries as values in a dictionary, keyed by cache key.
    This allows sharing evaluations across threads in a single process.

    The implementation is not thread-safe.

    Attributes:
        cached_values: Dictionary used to store cached values.

    ??? Examples
        ``` pycon
        >>> from pydvl.utils.caching.memory import InMemoryCacheBackend
        >>> cache_backend = InMemoryCacheBackend()
        >>> cache_backend.clear()
        >>> value = 42
        >>> cache_backend.set("key", value)
        >>> cache_backend.get("key")
        42
        ```

        ``` pycon
        >>> from pydvl.utils.caching.memory import InMemoryCacheBackend
        >>> cache_backend = InMemoryCacheBackend()
        >>> cache_backend.clear()
        >>> value = 42
        >>> def foo(x: int):
        ...     return x + 1
        ...
        >>> wrapped_foo = cache_backend.wrap(foo)
        >>> wrapped_foo(value)
        43
        >>> wrapped_foo.stats.misses
        1
        >>> wrapped_foo.stats.hits
        0
        >>> wrapped_foo(value)
        43
        >>> wrapped_foo.stats.misses
        1
        >>> wrapped_foo.stats.hits
        1
        ```
    """

    def __init__(self) -> None:
        """Initialize the in-memory cache backend."""
        super().__init__()
        self.cached_values: Dict[str, Any] = {}

    def get(self, key: str) -> Optional[Any]:
        """Get a value from the cache.

        Args:
            key: Cache key.

        Returns:
            Cached value or None if not found.
        """
        value = self.cached_values.get(key, None)
        if value is not None:
            self.stats.hits += 1
        else:
            self.stats.misses += 1
        return value

    def set(self, key: str, value: Any) -> None:
        """Set a value in the cache.

        Args:
            key: Cache key.
            value: Value to cache.
        """
        self.cached_values[key] = value
        self.stats.sets += 1

    def clear(self) -> None:
        """Deletes cache dictionary and recreates it."""
        del self.cached_values
        self.cached_values = {}

    def combine_hashes(self, *args: str) -> str:
        """Join cache key components."""
        return os.pathsep.join(args)



import itertools
import logging
import warnings
from typing import List, NamedTuple, Optional, Sequence, Tuple

import cvxpy as cp
import numpy as np
from numpy.typing import NDArray

from pydvl.parallel import MapReduceJob, ParallelConfig
from pydvl.utils import Status, Utility
from pydvl.value import ValuationResult

__all__ = [
    "_solve_least_core_linear_program",
    "_solve_egalitarian_least_core_quadratic_program",
    "lc_solve_problem",
    "lc_solve_problems",
    "LeastCoreProblem",
]

logger = logging.getLogger(__name__)


class LeastCoreProblem(NamedTuple):
    utility_values: NDArray[np.float_]
    A_lb: NDArray[np.float_]


def lc_solve_problem(
    problem: LeastCoreProblem,
    *,
    u: Utility,
    algorithm: str,
    non_negative_subsidy: bool = False,
    solver_options: Optional[dict] = None,
    **options,
) -> ValuationResult:
    """Solves a linear problem as prepared by
    [mclc_prepare_problem()][pydvl.value.least_core.montecarlo.mclc_prepare_problem].
    Useful for parallel execution of multiple experiments by running this as a
    remote task.

    See [exact_least_core()][pydvl.value.least_core.naive.exact_least_core] or
    [montecarlo_least_core()][pydvl.value.least_core.montecarlo.montecarlo_least_core] for
    argument descriptions.
    """
    n = len(u.data)

    if np.any(np.isnan(problem.utility_values)):
        warnings.warn(
            f"Calculation returned "
            f"{np.sum(np.isnan(problem.utility_values))} NaN "
            f"values out of {problem.utility_values.size}",
            RuntimeWarning,
        )

    # TODO: remove this before releasing version 0.7.0
    if options:
        warnings.warn(
            DeprecationWarning(
                "Passing solver options as kwargs was deprecated in "
                "0.6.0, will be removed in 0.7.0. `Use solver_options` "
                "instead."
            )
        )
        if solver_options is None:
            solver_options = options
        else:
            solver_options.update(options)

    if solver_options is None:
        solver_options = {}

    if "solver" not in solver_options:
        solver_options["solver"] = cp.SCS

    if "max_iters" not in solver_options and solver_options["solver"] == cp.SCS:
        solver_options["max_iters"] = 10000

    logger.debug("Removing possible duplicate values in lower bound array")
    b_lb = problem.utility_values
    A_lb, unique_indices = np.unique(problem.A_lb, return_index=True, axis=0)
    b_lb = b_lb[unique_indices]

    logger.debug("Building equality constraint")
    A_eq = np.ones((1, n))
    # We might have already computed the total utility one or more times.
    # This is the index of the row(s) in A_lb with all ones.
    total_utility_indices = np.where(A_lb.sum(axis=1) == n)[0]
    if len(total_utility_indices) == 0:
        b_eq = np.array([u(u.data.indices)])
    else:
        b_eq = b_lb[total_utility_indices]
        # Remove the row(s) corresponding to the total utility
        # from the lower bound constraints
        # because given the equality constraint
        # it is the same as using the constraint e >= 0
        # (i.e. setting non_negative_subsidy = True).
        mask: NDArray[np.bool_] = np.ones_like(b_lb, dtype=bool)
        mask[total_utility_indices] = False
        b_lb = b_lb[mask]
        A_lb = A_lb[mask]

    # Remove the row(s) corresponding to the empty subset
    # because, given u(∅) = (which is almost always the case,
    # it is the same as using the constraint e >= 0
    # (i.e. setting non_negative_subsidy = True).
    emptyset_utility_indices = np.where(A_lb.sum(axis=1) == 0)[0]
    if len(emptyset_utility_indices) > 0:
        mask = np.ones_like(b_lb, dtype=bool)
        mask[emptyset_utility_indices] = False
        b_lb = b_lb[mask]
        A_lb = A_lb[mask]

    _, subsidy = _solve_least_core_linear_program(
        A_eq=A_eq,
        b_eq=b_eq,
        A_lb=A_lb,
        b_lb=b_lb,
        non_negative_subsidy=non_negative_subsidy,
        solver_options=solver_options,
    )

    values: Optional[NDArray[np.float_]]

    if subsidy is None:
        logger.debug("No values were found")
        status = Status.Failed
        values = np.empty(n)
        values[:] = np.nan
        subsidy = np.nan
    else:
        values = _solve_egalitarian_least_core_quadratic_program(
            subsidy,
            A_eq=A_eq,
            b_eq=b_eq,
            A_lb=A_lb,
            b_lb=b_lb,
            solver_options=solver_options,
        )

        if values is None:
            logger.debug("No values were found")
            status = Status.Failed
            values = np.empty(n)
            values[:] = np.nan
            subsidy = np.nan
        else:
            status = Status.Converged

    return ValuationResult(
        algorithm=algorithm,
        status=status,
        values=values,
        subsidy=subsidy,
        stderr=None,
        data_names=u.data.data_names,
    )


def lc_solve_problems(
    problems: Sequence[LeastCoreProblem],
    u: Utility,
    algorithm: str,
    config: ParallelConfig = ParallelConfig(),
    n_jobs: int = 1,
    non_negative_subsidy: bool = True,
    solver_options: Optional[dict] = None,
    **options,
) -> List[ValuationResult]:
    """Solves a list of linear problems in parallel.

    Args:
        u: Utility.
        problems: Least Core problems to solve, as returned by
            [mclc_prepare_problem()][pydvl.value.least_core.montecarlo.mclc_prepare_problem].
        algorithm: Name of the valuation algorithm.
        config: Object configuring parallel computation, with cluster address,
            number of cpus, etc.
        n_jobs: Number of parallel jobs to run.
        non_negative_subsidy: If True, the least core subsidy $e$ is constrained
            to be non-negative.
        solver_options: Additional options to pass to the solver.

    Returns:
        List of solutions.
    """

    def _map_func(
        problems: List[LeastCoreProblem], *args, **kwargs
    ) -> List[ValuationResult]:
        return [lc_solve_problem(p, *args, **kwargs) for p in problems]

    map_reduce_job: MapReduceJob[
        "LeastCoreProblem", "List[ValuationResult]"
    ] = MapReduceJob(
        inputs=problems,
        map_func=_map_func,
        map_kwargs=dict(
            u=u,
            algorithm=algorithm,
            non_negative_subsidy=non_negative_subsidy,
            solver_options=solver_options,
            **options,
        ),
        reduce_func=lambda x: list(itertools.chain(*x)),
        config=config,
        n_jobs=n_jobs,
    )
    solutions = map_reduce_job()

    return solutions


def _solve_least_core_linear_program(
    A_eq: NDArray[np.float_],
    b_eq: NDArray[np.float_],
    A_lb: NDArray[np.float_],
    b_lb: NDArray[np.float_],
    solver_options: dict,
    non_negative_subsidy: bool = False,
) -> Tuple[Optional[NDArray[np.float_]], Optional[float]]:
    r"""Solves the Least Core's linear program using cvxopt.

    $$
        \text{minimize} \ & e \\
        \mbox{such that} \ & A_{eq} x = b_{eq}, \\
        & A_{lb} x + e \ge b_{lb},\\
        & A_{eq} x = b_{eq},\\
        & x in \mathcal{R}^n , \\
    $$
     where $x$ is a vector of decision variables; ,
    $b_{ub}$, $b_{eq}$, $l$, and $u$ are vectors; and
    $A_{ub}$ and $A_{eq}$ are matrices.

    if `non_negative_subsidy` is True, then an additional constraint $e \ge 0$ is used.

    Args:
        A_eq: The equality constraint matrix. Each row of `A_eq` specifies the
            coefficients of a linear equality constraint on `x`.
        b_eq: The equality constraint vector. Each element of `A_eq @ x` must equal
            the corresponding element of `b_eq`.
        A_lb: The inequality constraint matrix. Each row of `A_lb` specifies the
            coefficients of a linear inequality constraint on `x`.
        b_lb: The inequality constraint vector. Each element represents a
            lower bound on the corresponding value of `A_lb @ x`.
            non_negative_subsidy: If True, the least core subsidy $e$ is constrained
            to be non-negative.
        options: Keyword arguments that will be used to select a solver
            and to configure it. For all possible options, refer to [cvxpy's
            documentation](https://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options).
    """
    logger.debug(f"Solving linear program : {A_eq=}, {b_eq=}, {A_lb=}, {b_lb=}")

    n_variables = A_eq.shape[1]

    x = cp.Variable(n_variables)
    e = cp.Variable()

    objective = cp.Minimize(e)
    constraints = [A_eq @ x == b_eq, (A_lb @ x + e * np.ones(len(A_lb))) >= b_lb]

    if non_negative_subsidy:
        constraints += [e >= 0]

    problem = cp.Problem(objective, constraints)

    try:
        problem.solve(**solver_options)
    except cp.error.SolverError as err:
        raise ValueError("Could not solve linear program") from err

    if problem.status in cp.settings.SOLUTION_PRESENT:
        logger.debug("Problem was solved")
        if problem.status == cp.settings.USER_LIMIT:
            warnings.warn(
                "Solver terminated early. Consider increasing the solver's "
                "maximum number of iterations in solver_options",
                RuntimeWarning,
            )
        subsidy = e.value.item()
        return x.value, subsidy

    if problem.status in cp.settings.INF_OR_UNB:
        warnings.warn(
            "Could not find solution due to infeasibility or unboundedness of problem.",
            RuntimeWarning,
        )
    return None, None


def _solve_egalitarian_least_core_quadratic_program(
    subsidy: float,
    A_eq: NDArray[np.float_],
    b_eq: NDArray[np.float_],
    A_lb: NDArray[np.float_],
    b_lb: NDArray[np.float_],
    solver_options: dict,
) -> Optional[NDArray[np.float_]]:
    r"""Solves the egalitarian Least Core's quadratic program using cvxopt.

    $$
        \text{minimize} \ & \| x \|_2 \\
        \mbox{such that} \ & A_{eq} x = b_{eq}, \\
        & A_{lb} x + e \ge b_{lb},\\
        & A_{eq} x = b_{eq},\\
        & x in \mathcal{R}^n , \\
        & e \text{ is a constant.}
    $$
     where $x$ is a vector of decision variables; ,
    $b_{ub}$, $b_{eq}$, $l$, and $u$ are vectors; and
    $A_{ub}$ and $A_{eq}$ are matrices.

    Args:
        subsidy: Minimal subsidy returned by
            [_solve_least_core_linear_program()][pydvl.value.least_core.common._solve_least_core_linear_program]
        A_eq: The equality constraint matrix. Each row of `A_eq` specifies the
            coefficients of a linear equality constraint on `x`.
        b_eq: The equality constraint vector. Each element of `A_eq @ x` must equal
            the corresponding element of `b_eq`.
        A_lb: The inequality constraint matrix. Each row of `A_lb` specifies the
            coefficients of a linear inequality constraint on `x`.
        b_lb: The inequality constraint vector. Each element represents a
            lower bound on the corresponding value of `A_lb @ x`.
        solver_options: Keyword arguments that will be used to select a solver
            and to configure it. Refer to [cvxpy's
            documentation](https://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options)
            for all possible options.
    """
    logger.debug(f"Solving quadratic program : {A_eq=}, {b_eq=}, {A_lb=}, {b_lb=}")

    n_variables = A_eq.shape[1]

    x = cp.Variable(n_variables)

    objective = cp.Minimize(cp.norm2(x))
    constraints = [A_eq @ x == b_eq, (A_lb @ x + subsidy * np.ones(len(A_lb))) >= b_lb]
    problem = cp.Problem(objective, constraints)

    try:
        problem.solve(**solver_options)
    except cp.error.SolverError as err:
        raise ValueError("Could not solve quadratic program") from err

    if problem.status in cp.settings.SOLUTION_PRESENT:
        logger.debug("Problem was solved")
        if problem.status == cp.settings.USER_LIMIT:
            warnings.warn(
                "Solver terminated early. Consider increasing the solver's "
                "maximum number of iterations in solver_options",
                RuntimeWarning,
            )
        return x.value  # type: ignore

    if problem.status in cp.settings.INF_OR_UNB:
        warnings.warn(
            "Could not find solution due to infeasibility or unboundedness of problem.",
            RuntimeWarning,
        )
    return None



import os
import shutil
import tempfile
from pathlib import Path
from typing import Any, Optional, Union

import cloudpickle

from pydvl.utils.caching.base import CacheBackend

__all__ = ["DiskCacheBackend"]

PICKLE_VERSION = 5  # python >= 3.8


class DiskCacheBackend(CacheBackend):
    """Disk cache backend that stores results in files.

    Implements the CacheBackend interface for a disk-based cache.
    Stores cache entries as pickled files on disk, keyed by cache key.
    This allows sharing evaluations across processes in a single node/computer.

    Args:
        cache_dir: Base directory for cache storage.

    Attributes:
        cache_dir: Base directory for cache storage.

    ??? Examples
        ``` pycon
        >>> from pydvl.utils.caching.disk import DiskCacheBackend
        >>> cache_backend = DiskCacheBackend()
        >>> cache_backend.clear()
        >>> value = 42
        >>> cache_backend.set("key", value)
        >>> cache_backend.get("key")
        42
        ```

        ``` pycon
        >>> from pydvl.utils.caching.disk import DiskCacheBackend
        >>> cache_backend = DiskCacheBackend()
        >>> cache_backend.clear()
        >>> value = 42
        >>> def foo(x: int):
        ...     return x + 1
        ...
        >>> wrapped_foo = cache_backend.wrap(foo)
        >>> wrapped_foo(value)
        43
        >>> wrapped_foo.stats.misses
        1
        >>> wrapped_foo.stats.hits
        0
        >>> wrapped_foo(value)
        43
        >>> wrapped_foo.stats.misses
        1
        >>> wrapped_foo.stats.hits
        1
        ```

    """

    def __init__(
        self,
        cache_dir: Optional[Union[os.PathLike, str]] = None,
    ) -> None:
        """Initialize the disk cache backend.

        Args:
            cache_dir: Base directory for cache storage.
                If not provided, this defaults to a newly created
                temporary directory.
        """
        super().__init__()
        if cache_dir is None:
            cache_dir = tempfile.mkdtemp(prefix="pydvl")
        self.cache_dir = Path(cache_dir)
        self.cache_dir.mkdir(exist_ok=True, parents=True)

    def get(self, key: str) -> Optional[Any]:
        """Get a value from the cache.

        Args:
            key: Cache key.

        Returns:
            Cached value or None if not found.
        """
        cache_file = self.cache_dir / key
        if not cache_file.exists():
            self.stats.misses += 1
            return None
        self.stats.hits += 1
        with cache_file.open("rb") as f:
            return cloudpickle.load(f)

    def set(self, key: str, value: Any) -> None:
        """Set a value in the cache.

        Args:
            key: Cache key.
            value: Value to cache.
        """
        cache_file = self.cache_dir / key
        self.stats.sets += 1
        with cache_file.open("wb") as f:
            cloudpickle.dump(value, f, protocol=PICKLE_VERSION)

    def clear(self) -> None:
        """Deletes cache directory and recreates it."""
        shutil.rmtree(self.cache_dir)
        self.cache_dir.mkdir(exist_ok=True, parents=True)

    def combine_hashes(self, *args: str) -> str:
        """Join cache key components."""
        return os.pathsep.join(args)



import logging
import math
from functools import partial
from typing import (
    Collection,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

import dask
import numpy as np
import torch
from dask import array as da
from numpy.typing import NDArray
from torch.utils.data import Dataset

from ..array import NestedSequenceAggregator, NumpyConverter, SequenceAggregator

logger = logging.getLogger(__name__)

__all__ = [
    "to_model_device",
    "TorchTensorContainerType",
    "align_structure",
    "align_with_model",
    "flatten_dimensions",
    "TorchNumpyConverter",
    "TorchCatAggregator",
    "NestedTorchCatAggregator",
    "torch_dataset_to_dask_array",
]


def to_model_device(x: torch.Tensor, model: torch.nn.Module) -> torch.Tensor:
    """
    Returns the tensor `x` moved to the device of the `model`, if device of model is set

    Args:
        x: The tensor to be moved to the device of the model.
        model: The model whose device will be used to move the tensor.

    Returns:
        The tensor `x` moved to the device of the `model`, if device of model is set.
    """
    device = next(model.parameters()).device
    return x.to(device)


def reshape_vector_to_tensors(
    input_vector: torch.Tensor, target_shapes: Iterable[Tuple[int, ...]]
) -> Tuple[torch.Tensor, ...]:
    """
    Reshape a 1D tensor into multiple tensors with specified shapes.

    This function takes a 1D tensor (input_vector) and reshapes it into a series of
    tensors with shapes given by 'target_shapes'.
    The reshaped tensors are returned as a tuple in the same order
    as their corresponding shapes.

    Note:
        The total number of elements in 'input_vector' must be equal to the
            sum of the products of the shapes in 'target_shapes'.

    Args:
        input_vector: The 1D tensor to be reshaped. Must be 1D.
        target_shapes: An iterable of tuples. Each tuple defines the shape of a tensor
            to be reshaped from the 'input_vector'.

    Returns:
        A tuple of reshaped tensors.

    Raises:
        ValueError: If 'input_vector' is not a 1D tensor or if the total
            number of elements in 'input_vector' does not
            match the sum of the products of the shapes in 'target_shapes'.
    """

    if input_vector.dim() != 1:
        raise ValueError("Input vector must be a 1D tensor")

    total_elements = sum(math.prod(shape) for shape in target_shapes)

    if total_elements != input_vector.shape[0]:
        raise ValueError(
            f"The total elements in shapes {total_elements} "
            f"does not match the vector length {input_vector.shape[0]}"
        )

    tensors = []
    start = 0
    for shape in target_shapes:
        size = math.prod(shape)  # compute the total size of the tensor with this shape
        tensors.append(
            input_vector[start : start + size].view(shape)
        )  # slice the vector and reshape it
        start += size
    return tuple(tensors)


TorchTensorContainerType = Union[
    torch.Tensor,
    Collection[torch.Tensor],
    Mapping[str, torch.Tensor],
]
"""Type for a PyTorch tensor or a container thereof."""


def align_structure(
    source: Mapping[str, torch.Tensor],
    target: TorchTensorContainerType,
) -> Dict[str, torch.Tensor]:
    """
    This function transforms `target` to have the same structure as `source`, i.e.,
    it should be a dictionary with the same keys as `source` and each corresponding
    value in `target` should have the same shape as the value in `source`.

    Args:
        source: The reference dictionary containing PyTorch tensors.
        target: The input to be harmonized. It can be a dictionary, tuple, or tensor.

    Returns:
        The harmonized version of `target`.

    Raises:
        ValueError: If `target` cannot be harmonized to match `source`.
    """

    tangent_dict: Dict[str, torch.Tensor]

    if isinstance(target, dict):

        if list(target.keys()) != list(source.keys()):
            raise ValueError("The keys in 'target' do not match the keys in 'source'.")

        if [v.shape for v in target.values()] != [v.shape for v in source.values()]:

            raise ValueError(
                "The shapes of the values in 'target' do not match the shapes "
                "of the values in 'source'."
            )

        tangent_dict = target

    elif isinstance(target, tuple) or isinstance(target, list):

        if [v.shape for v in target] != [v.shape for v in source.values()]:

            raise ValueError(
                "'target' is a tuple/list but its elements' shapes do not match "
                "the shapes of the values in 'source'."
            )

        tangent_dict = dict(zip(source.keys(), target))

    elif isinstance(target, torch.Tensor):

        try:
            tangent_dict = dict(
                zip(
                    source.keys(),
                    reshape_vector_to_tensors(
                        target, [p.shape for p in source.values()]
                    ),
                )
            )
        except Exception as e:
            raise ValueError(
                f"'target' is a tensor but cannot be reshaped to match 'source'. "
                f"Original error: {e}"
            )

    else:
        raise ValueError(f"'target' is of type {type(target)} which is not supported.")

    return tangent_dict


def align_with_model(x: TorchTensorContainerType, model: torch.nn.Module):
    """
    Aligns an input to the model's parameter structure, i.e. transforms it into a dict
    with the same keys as model.named_parameters() and matching tensor shapes

    Args:
        x: The input to be aligned. It can be a dictionary, tuple, or tensor.
        model: model to use for alignment

    Returns:
        The aligned version of `x`.

    Raises:
        ValueError: If `x` cannot be aligned to match the model's parameters .

    """
    model_params = {k: p for k, p in model.named_parameters() if p.requires_grad}
    return align_structure(model_params, x)


def flatten_dimensions(
    tensors: Iterable[torch.Tensor],
    shape: Optional[Tuple[int, ...]] = None,
    concat_at: int = -1,
) -> torch.Tensor:
    """
    Flattens the dimensions of each tensor in the given iterable and concatenates them
    along a specified dimension.

    This function takes an iterable of PyTorch tensors and flattens each tensor.
    Optionally, each tensor can be reshaped to a specified shape before concatenation.
    The concatenation is performed along the dimension specified by `concat_at`.

    Args:
        tensors: An iterable containing PyTorch tensors to be flattened
            and concatenated.
        shape: A tuple representing the desired shape to which each tensor is reshaped
            before concatenation. If None, tensors are flattened to 1D.
        concat_at: The dimension along which to concatenate the tensors.

    Returns:
        A single tensor resulting from the concatenation of the input tensors,
        each either flattened or reshaped as specified.

    ??? Example
        ```pycon
        >>> tensors = [torch.tensor([[1, 2], [3, 4]]), torch.tensor([[5, 6], [7, 8]])]
        >>> flatten_dimensions(tensors)
        tensor([1, 2, 3, 4, 5, 6, 7, 8])

        >>> flatten_dimensions(tensors, shape=(2, 2), concat_at=0)
        tensor([[1, 2],
                [3, 4],
                [5, 6],
                [7, 8]])
        ```
    """
    return torch.cat(
        [t.reshape(-1) if shape is None else t.reshape(*shape) for t in tensors],
        dim=concat_at,
    )


def torch_dataset_to_dask_array(
    dataset: Dataset,
    chunk_size: int,
    total_size: Optional[int] = None,
    resulting_dtype=np.float32,
) -> Tuple[da.Array, ...]:
    """
    Construct tuple of dask arrays from a PyTorch dataset, using dask.delayed

    Args:
        dataset: A PyTorch [dataset][torch.utils.data.Dataset]
        chunk_size: The size of the chunks for the resulting Dask arrays.
        total_size: If the dataset does not implement len, provide the length
            via this parameter. If None
            the length of the dataset is inferred via accessing the dataset once.
        resulting_dtype: The dtype of the resulting [dask.array.Array][dask.array.Array]

    ??? Example
        ```python
        import torch
        from torch.utils.data import TensorDataset
        x = torch.rand((20, 3))
        y = torch.rand((20, 1))
        dataset = TensorDataset(x, y)
        da_x, da_y = torch_dataset_to_dask_array(dataset, 4)
        ```

    Returns:
        Tuple of Dask arrays corresponding to each tensor in the dataset.
    """

    def _infer_data_len(d_set: Dataset):
        try:
            n_data = len(d_set)
            if total_size is not None and n_data != total_size:
                raise ValueError(
                    f"The number of samples in the dataset ({n_data}), derived "
                    f"from calling ´len´, does not match the provided "
                    f"total number of samples ({total_size}). "
                    f"Call the function without total_size."
                )
            return n_data
        except TypeError as e:
            err_msg = (
                f"Could not infer the number of samples in the dataset from "
                f"calling ´len´. Original error: {e}."
            )
            if total_size is not None:
                logger.warning(
                    err_msg
                    + f" Using the provided total number of samples {total_size}."
                )
                return total_size
            else:
                logger.warning(
                    err_msg + f" Infer the number of samples from the dataset, "
                    f"via iterating the dataset once. "
                    f"This might induce severe overhead, so consider"
                    f"providing total_size, if you know the number of samples "
                    f"beforehand."
                )
                idx = 0
                while True:
                    try:
                        t = d_set[idx]
                        if all(_t.numel() == 0 for _t in t):
                            return idx
                        idx += 1

                    except IndexError:
                        return idx

    sample = dataset[0]
    if not isinstance(sample, tuple):
        sample = (sample,)

    def _get_chunk(
        start_idx: int, stop_idx: int, d_set: Dataset
    ) -> Tuple[torch.Tensor, ...]:
        try:
            t = d_set[start_idx:stop_idx]
            if not isinstance(t, tuple):
                t = (t,)
            return t  # type:ignore
        except Exception:
            nested_tensor_list = [
                [d_set[idx][k] for idx in range(start_idx, stop_idx)]
                for k in range(len(sample))
            ]
            return tuple(map(torch.stack, nested_tensor_list))

    n_samples = _infer_data_len(dataset)
    chunk_indices = [
        (i, min(i + chunk_size, n_samples)) for i in range(0, n_samples, chunk_size)
    ]
    delayed_dataset = dask.delayed(dataset)
    delayed_chunks = [
        dask.delayed(partial(_get_chunk, start, stop))(delayed_dataset)
        for (start, stop) in chunk_indices
    ]

    delayed_arrays_dict: Dict[int, List[da.Array]] = {k: [] for k in range(len(sample))}

    for chunk, (start, stop) in zip(delayed_chunks, chunk_indices):
        for tensor_idx, sample_tensor in enumerate(sample):

            delayed_tensor = da.from_delayed(
                dask.delayed(lambda t: t.cpu().numpy())(chunk[tensor_idx]),
                shape=(stop - start, *sample_tensor.shape),
                dtype=resulting_dtype,
            )

            delayed_arrays_dict[tensor_idx].append(delayed_tensor)

    return tuple(
        da.concatenate(array_list) for array_list in delayed_arrays_dict.values()
    )


class TorchNumpyConverter(NumpyConverter[torch.Tensor]):
    """
    Helper class for converting between [torch.Tensor][torch.Tensor] and
    [numpy.ndarray][numpy.ndarray]

    Args:
        device: Optional device parameter to move the resulting torch tensors to the
            specified device

    """

    def __init__(self, device: Optional[torch.device] = None):
        self.device = device

    def to_numpy(self, x: torch.Tensor) -> NDArray:
        """
        Convert a detached [torch.Tensor][torch.Tensor] to
        [numpy.ndarray][numpy.ndarray]
        """
        arr: NDArray = x.cpu().numpy()
        return arr

    def from_numpy(self, x: NDArray) -> torch.Tensor:
        """
        Convert a [numpy.ndarray][numpy.ndarray] to [torch.Tensor][torch.Tensor] and
        optionally move it to a provided device
        """
        t = torch.from_numpy(x)
        if self.device is not None:
            t = t.to(self.device)
        return t


class TorchCatAggregator(SequenceAggregator[torch.Tensor]):
    """
    An aggregator that concatenates tensors using PyTorch's [torch.cat][torch.cat]
    function. Concatenation is done along the first dimension of the chunks.
    """

    def __call__(self, tensor_generator: Generator[torch.Tensor, None, None]):
        """
        Aggregates tensors from a single-level generator into a single tensor by
        concatenating them. This method is a straightforward way to combine a sequence
        of tensors into one larger tensor.

        Args:
            tensor_generator: A generator that yields `torch.Tensor` objects.

        Returns:
            A single tensor formed by concatenating all tensors from the generator.
                The concatenation is performed along the default dimension (0).
        """
        return torch.cat(list(tensor_generator))


class NestedTorchCatAggregator(NestedSequenceAggregator[torch.Tensor]):
    """
    An aggregator that concatenates tensors using PyTorch's [torch.cat][torch.cat]
    function. Concatenation is done along the first two dimensions of the chunks.
    """

    def __call__(
        self,
        nested_generators_of_tensors: Generator[
            Generator[torch.Tensor, None, None], None, None
        ],
    ):
        """
        Aggregates tensors from a nested generator structure into a single tensor by
        concatenating. Each inner generator is first concatenated along dimension 1 into
        a tensor, and then these tensors are concatenated along dimension 0 together to
        form the final tensor.

        Args:
            nested_generators_of_tensors: A generator of generators, where each inner
                generator yields `torch.Tensor` objects.

        Returns:
            A single tensor formed by concatenating all tensors from the nested
            generators.

        """
        return torch.cat(
            list(
                map(
                    lambda tensor_gen: torch.cat(list(tensor_gen), dim=1),
                    nested_generators_of_tensors,
                )
            )
        )



from abc import ABC, abstractmethod
from enum import Enum
from typing import Collection, Generic, Iterable, Optional, TypeVar


class InfluenceMode(str, Enum):
    r"""
    Enum representation for the types of influence.

    Attributes:
        Up: [Approximating the influence of a point]
            [approximating-the-influence-of-a-point]
        Perturbation: [Perturbation definition of the influence score]
            [perturbation-definition-of-the-influence-score]

    """

    Up = "up"
    Perturbation = "perturbation"


class UnsupportedInfluenceModeException(ValueError):
    def __init__(self, mode: str):
        super().__init__(
            f"Provided {mode=} is not supported. Choose one of InfluenceMode.Up "
            f"and InfluenceMode.Perturbation"
        )


class NotFittedException(ValueError):
    def __init__(self):
        super().__init__(
            f"Objects of type InfluenceFunctionModel must be fitted before calling "
            f"influence methods. "
            f"Call method fit with appropriate input."
        )


"""Type variable for tensors, i.e. sequences of numbers"""
TensorType = TypeVar("TensorType", bound=Collection)
DataLoaderType = TypeVar("DataLoaderType", bound=Iterable)


class InfluenceFunctionModel(Generic[TensorType, DataLoaderType], ABC):
    """
    Generic abstract base class for computing influence related quantities.
    For a specific influence algorithm and tensor framework,
    inherit from this base class
    """

    @property
    @abstractmethod
    def n_parameters(self):
        """Number of trainable parameters of the underlying model"""

    @property
    @abstractmethod
    def is_thread_safe(self) -> bool:
        """Whether the influence computation is thread safe"""

    @property
    @abstractmethod
    def is_fitted(self):
        """Override this, to expose the fitting status of the instance."""

    @abstractmethod
    def fit(self, data: DataLoaderType):
        """
        Override this method to fit the influence function model to training data,
        e.g. pre-compute hessian matrix or matrix decompositions

        Args:
            data:

        Returns:
            The fitted instance
        """

    def influence_factors(self, x: TensorType, y: TensorType) -> TensorType:
        if not self.is_fitted:
            raise NotFittedException()
        return self._influence_factors(x, y)

    @abstractmethod
    def _influence_factors(self, x: TensorType, y: TensorType) -> TensorType:
        r"""
        Override this method to implement the approximation of

        \[ H^{-1}\nabla_{\theta} \ell(y, f_{\theta}(x)) \]

        where the gradient is meant to be per sample of the batch $(x, y)$.

        Args:
            x: model input to use in the gradient computations
            y: label tensor to compute gradients

        Returns:
            Tensor representing the element-wise inverse Hessian matrix vector products

        """

    def influences(
        self,
        x_test: TensorType,
        y_test: TensorType,
        x: Optional[TensorType] = None,
        y: Optional[TensorType] = None,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> TensorType:
        if not self.is_fitted:
            raise NotFittedException()
        return self._influences(x_test, y_test, x, y, mode)

    @abstractmethod
    def _influences(
        self,
        x_test: TensorType,
        y_test: TensorType,
        x: Optional[TensorType] = None,
        y: Optional[TensorType] = None,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> TensorType:
        r"""
        Override this method to implement the approximation of

        \[ \langle H^{-1}\nabla_{theta} \ell(y_{\text{test}},
            f_{\theta}(x_{\text{test}})),
            \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the case of up-weighting influence, resp.

        \[ \langle H^{-1}\nabla_{theta} \ell(y_{test}, f_{\theta}(x_{test})),
            \nabla_{x} \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the perturbation type influence case.

        Args:
            x_test: model input to use in the gradient computations
                of $H^{-1}\nabla_{theta} \ell(y_{test}, f_{\theta}(x_{test}))$
            y_test: label tensor to compute gradients
            x: optional model input to use in the gradient computations
                $\nabla_{theta}\ell(y, f_{\theta}(x))$,
                resp. $\nabla_{x}\nabla_{theta}\ell(y, f_{\theta}(x))$,
                if None, use $x=x_{test}$
            y: optional label tensor to compute gradients
            mode: enum value of [InfluenceMode]
                [pydvl.influence.base_influence_modl.InfluenceMode]

        Returns:
            Tensor representing the element-wise scalar products for the provided batch

        """

    @abstractmethod
    def influences_from_factors(
        self,
        z_test_factors: TensorType,
        x: TensorType,
        y: TensorType,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> TensorType:
        r"""
        Override this method to implement the computation of

        \[ \langle z_{\text{test_factors}},
            \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the case of up-weighting influence, resp.

        \[ \langle z_{\text{test_factors}},
            \nabla_{x} \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the perturbation type influence case. The gradient is meant to be per sample
        of the batch $(x, y)$.

        Args:
             z_test_factors: pre-computed array, approximating
                $H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
                f_{\theta}(x_{\text{test}}))$
             x: model input to use in the gradient computations
                $\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                resp. $\nabla_{x}\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                if None, use $x=x_{\text{test}}$
             y: label tensor to compute gradients
             mode: enum value of [InfluenceMode]
                [pydvl.influence.base_influence_modl.InfluenceMode]

        Returns:
            Tensor representing the element-wise scalar products for the provided batch

        """



import logging
from functools import wraps
from itertools import cycle, takewhile
from time import time
from typing import TYPE_CHECKING, Collection, Iterator

from tqdm.auto import tqdm

# This is needed to avoid circular import errors
if TYPE_CHECKING:
    from pydvl.value.result import ValuationResult
    from pydvl.value.stopping import StoppingCriterion

__all__ = ["repeat_indices", "log_duration"]

logger = logging.getLogger(__name__)


def repeat_indices(
    indices: Collection[int],
    result: "ValuationResult",
    done: "StoppingCriterion",
    **kwargs,
) -> Iterator[int]:
    """Helper function to cycle indefinitely over a collection of indices
    until the stopping criterion is satisfied while displaying progress.

    Args:
        indices: Collection of indices that will be cycled until done.
        result: Object containing the current results.
        done: Stopping criterion.
        kwargs: Keyword arguments passed to tqdm.
    """
    with tqdm(total=100, unit="%", **kwargs) as pbar:
        it = takewhile(lambda _: not done(result), cycle(indices))
        for i in it:
            yield i
            pbar.update(100 * done.completion() - pbar.n)
            pbar.refresh()


def log_duration(func):
    """
    Decorator to log execution time of a function
    """

    @wraps(func)
    def wrapper_log_duration(*args, **kwargs):
        func_name = func.__qualname__
        logger.info(f"Function '{func_name}' is starting.")
        start_time = time()
        result = func(*args, **kwargs)
        duration = time() - start_time
        logger.info(f"Function '{func_name}' completed. Duration: {duration:.2f} sec")
        return result

    return wrapper_log_duration



"""
This package holds all routines for the computation of Shapley Data value. Users
will want to use
[compute_shapley_values][pydvl.value.shapley.common.compute_shapley_values] or
[compute_semivalues][pydvl.value.semivalues.compute_semivalues] as interfaces to
most methods defined in the modules.

Please refer to [the guide on data valuation][data-valuation] for an overview of
all methods.
"""

from ..result import *
from ..stopping import *
from .classwise import *
from .common import *
from .gt import *
from .knn import *
from .montecarlo import *
from .naive import *
from .owen import *
from .truncated import *
from .types import *



from functools import partial
from typing import Any, List, Literal, Optional, OrderedDict, Sequence

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy as sp
from deprecate import deprecated
from matplotlib.axes import Axes
from numpy.typing import NDArray
from scipy.stats import norm, t

from pydvl.value import ValuationResult


@deprecated(target=None, deprecated_in="0.7.1", remove_in="0.9.0")
def shaded_mean_std(
    data: np.ndarray,
    abscissa: Optional[Sequence[Any]] = None,
    num_std: float = 1.0,
    mean_color: Optional[str] = "dodgerblue",
    shade_color: Optional[str] = "lightblue",
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    ax: Optional[Axes] = None,
    **kwargs,
) -> Axes:
    r"""The usual mean \(\pm\) std deviation plot to aggregate runs of
    experiments.

    !!! warning "Deprecation notice"
        This function is bogus and will be removed in the future in favour of
        properly computed confidence intervals.

    Args:
        data: axis 0 is to be aggregated on (e.g. runs) and axis 1 is the
            data for each run.
        abscissa: values for the x-axis. Leave empty to use increasing integers.
        num_std: number of standard deviations to shade around the mean.
        mean_color: color for the mean
        shade_color: color for the shaded region
        title: Title text. To use mathematics, use LaTeX notation.
        xlabel: Text for the horizontal axis.
        ylabel: Text for the vertical axis
        ax: If passed, axes object into which to insert the figure. Otherwise,
            a new figure is created and returned
        kwargs: these are forwarded to the ax.plot() call for the mean.

    Returns:
        The axes used (or created)
    """
    assert len(data.shape) == 2
    mean = data.mean(axis=0)
    std = num_std * data.std(axis=0)

    if ax is None:
        fig, ax = plt.subplots()
    if abscissa is None:
        abscissa = list(range(data.shape[1]))

    ax.fill_between(abscissa, mean - std, mean + std, alpha=0.3, color=shade_color)
    ax.plot(abscissa, mean, color=mean_color, **kwargs)

    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return ax


def plot_ci_array(
    data: NDArray,
    level: float,
    type: Literal["normal", "t", "auto"] = "normal",
    abscissa: Optional[Sequence[str]] = None,
    mean_color: Optional[str] = "dodgerblue",
    shade_color: Optional[str] = "lightblue",
    ax: Optional[plt.Axes] = None,
    **kwargs,
) -> plt.Axes:
    """Plot values and a confidence interval from a 2D array.

    Supported intervals are based on the normal and the t distributions.

    Args:
        data: A 2D array with M different values for each of the N indices.
        level: The confidence level.
        type: The type of confidence interval to use.
        abscissa: The values for the x-axis. Leave empty to use increasing
            integers.
        mean_color: The color of the mean line.
        shade_color: The color of the confidence interval.
        ax: If passed, axes object into which to insert the figure. Otherwise,
            a new figure is created and the axes returned.
        **kwargs: Additional arguments to pass to the plot function.

    Returns:
        The matplotlib axes.
    """

    m, n = data.shape

    means = np.mean(data, axis=0)
    variances = np.var(data, axis=0, ddof=1)

    dummy = ValuationResult[np.int_, np.object_](
        algorithm="dummy",
        values=means,
        variances=variances,
        counts=np.ones_like(means, dtype=np.int_) * m,
        indices=np.arange(n),
        data_names=np.array(abscissa, dtype=str)
        if abscissa is not None
        else np.arange(n, dtype=str),
    )

    return plot_ci_values(
        dummy,
        level=level,
        type=type,
        mean_color=mean_color,
        shade_color=shade_color,
        ax=ax,
        **kwargs,
    )


def plot_ci_values(
    values: ValuationResult,
    level: float,
    type: Literal["normal", "t", "auto"] = "auto",
    abscissa: Optional[Sequence[str]] = None,
    mean_color: Optional[str] = "dodgerblue",
    shade_color: Optional[str] = "lightblue",
    ax: Optional[plt.Axes] = None,
    **kwargs,
):
    """Plot values and a confidence interval.

    Uses `values.data_names` for the x-axis.

    Supported intervals are based on the normal and the t distributions.

    Args:
        values: The valuation result.
        level: The confidence level.
        type: The type of confidence interval to use. If "auto", uses "norm" if
            the minimum number of updates for all indices is greater than 30,
            otherwise uses "t".
        abscissa: The values for the x-axis. Leave empty to use increasing
            integers.
        mean_color: The color of the mean line.
        shade_color: The color of the confidence interval.
        ax: If passed, axes object into which to insert the figure. Otherwise,
            a new figure is created and the axes returned.
        **kwargs: Additional arguments to pass to the plot function.

    Returns:
        The matplotlib axes.
    """

    ppfs = {
        "normal": norm.ppf,
        "t": partial(t.ppf, df=values.counts - 1),
        "auto": norm.ppf
        if np.min(values.counts) > 30
        else partial(t.ppf, df=values.counts - 1),
    }

    try:
        score = ppfs[type](1 - level / 2)
    except KeyError:
        raise ValueError(
            f"Unknown confidence interval type requested: {type}."
        ) from None

    if abscissa is None:
        abscissa = [str(i) for i, _ in enumerate(values)]
    bound = score * values.stderr

    if ax is None:
        fig, ax = plt.subplots()

    ax.fill_between(
        abscissa,
        values.values - bound,
        values.values + bound,
        alpha=0.3,
        color=shade_color,
    )
    ax.plot(abscissa, values.values, color=mean_color, **kwargs)
    return ax


def spearman_correlation(vv: List[OrderedDict], num_values: int, pvalue: float):
    """Simple matrix plots with spearman correlation for each pair in vv.

    Args:
        vv: list of OrderedDicts with index: value. Spearman correlation
            is computed for the keys.
        num_values: Use only these many values from the data (from the start
            of the OrderedDicts)
        pvalue: correlation coefficients for which the p-value is below the
            threshold `pvalue/len(vv)` will be discarded.
    """
    r: np.ndarray = np.ndarray((len(vv), len(vv)))
    p: np.ndarray = np.ndarray((len(vv), len(vv)))
    for i, a in enumerate(vv):
        for j, b in enumerate(vv):
            from scipy.stats._stats_py import SpearmanrResult

            spearman: SpearmanrResult = sp.stats.spearmanr(
                list(a.keys())[:num_values], list(b.keys())[:num_values]
            )
            r[i][j] = (
                spearman.correlation if spearman.pvalue < pvalue / len(vv) else np.nan
            )  # Bonferroni correction
            p[i][j] = spearman.pvalue
    fig, axs = plt.subplots(1, 2, figsize=(16, 7))
    plot1 = axs[0].matshow(r, vmin=-1, vmax=1)
    axs[0].set_title(f"Spearman correlation (top {num_values} values)")
    axs[0].set_xlabel("Runs")
    axs[0].set_ylabel("Runs")
    fig.colorbar(plot1, ax=axs[0])
    plot2 = axs[1].matshow(p, vmin=0, vmax=1)
    axs[1].set_title("p-value")
    axs[1].set_xlabel("Runs")
    axs[1].set_ylabel("Runs")
    fig.colorbar(plot2, ax=axs[1])

    return fig


def plot_shapley(
    df: pd.DataFrame,
    *,
    level: float = 0.05,
    ax: Optional[plt.Axes] = None,
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
) -> plt.Axes:
    r"""Plots the shapley values, as returned from
    [compute_shapley_values][pydvl.value.shapley.common.compute_shapley_values],
    with error bars corresponding to an $\alpha$-level Normal confidence
    interval.

    Args:
        df: dataframe with the shapley values
        level: confidence level for the error bars
        ax: axes to plot on or None if a new subplots should be created
        title: string, title of the plot
        xlabel: string, x label of the plot
        ylabel: string, y label of the plot

    Returns:
        The axes created or used
    """
    if ax is None:
        _, ax = plt.subplots()

    yerr = norm.ppf(1 - level / 2) * df["data_value_stderr"]

    ax.errorbar(x=df.index, y=df["data_value"], yerr=yerr, fmt="o", capsize=6)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    plt.xticks(rotation=60)
    return ax


def plot_influence_distribution(
    influences: NDArray[np.float_], index: int, title_extra: str = ""
) -> plt.Axes:
    """Plots the histogram of the influence that all samples in the training set
    have over a single sample index.

    Args:
       influences: array of influences (training samples x test samples)
       index: Index of the test sample for which the influences
            will be plotted.
       title_extra: Additional text that will be appended to the title.
    """
    _, ax = plt.subplots()
    ax.hist(influences[:, index], alpha=0.7)
    ax.set_xlabel("Influence values")
    ax.set_ylabel("Number of samples")
    ax.set_title(f"Distribution of influences {title_extra}")
    return ax


def plot_influence_distribution_by_label(
    influences: NDArray[np.float_], labels: NDArray[np.float_], title_extra: str = ""
):
    """Plots the histogram of the influence that all samples in the training set
    have over a single sample index, separated by labels.

    Args:
       influences: array of influences (training samples x test samples)
       labels: labels for the training set.
       title_extra: Additional text that will be appended to the title.
    """
    _, ax = plt.subplots()
    unique_labels = np.unique(labels)
    for label in unique_labels:
        ax.hist(influences[labels == label], label=label, alpha=0.7)
    ax.set_xlabel("Influence values")
    ax.set_ylabel("Number of samples")
    ax.set_title(f"Distribution of influences {title_extra}")
    ax.legend()
    plt.show()



import math
import warnings
from itertools import permutations
from typing import List

import numpy as np
from numpy.typing import NDArray
from tqdm.auto import tqdm

from pydvl.parallel import MapReduceJob, ParallelConfig
from pydvl.utils import Utility, powerset
from pydvl.utils.status import Status
from pydvl.value.result import ValuationResult

__all__ = ["permutation_exact_shapley", "combinatorial_exact_shapley"]


def permutation_exact_shapley(u: Utility, *, progress: bool = True) -> ValuationResult:
    r"""Computes the exact Shapley value using the formulation with permutations:

    $$v_u(x_i) = \frac{1}{n!} \sum_{\sigma \in \Pi(n)} [u(\sigma_{i-1} \cup {i}) − u(\sigma_{i})].$$

    See [Data valuation][computing-data-values] for details.

    When the length of the training set is > 10 this prints a warning since the
    computation becomes too expensive. Used mostly for internal testing and
    simple use cases. Please refer to the [Monte Carlo
    approximations][pydvl.value.shapley.montecarlo] for practical applications.

    Args:
        u: Utility object with model, data, and scoring function
        progress: Whether to display progress bars for each job.

    Returns:
        Object with the data values.
    """

    n = len(u.data)
    # Note that the cache in utility saves most of the refitting because we
    # use frozenset for the input.
    if n > 10:
        warnings.warn(
            f"Large dataset! Computation requires {n}! calls to utility()",
            RuntimeWarning,
        )

    values = np.zeros(n)
    for p in tqdm(
        permutations(u.data.indices),
        disable=not progress,
        desc="Permutation",
        total=math.factorial(n),
    ):
        for i, idx in enumerate(p):
            values[idx] += u(p[: i + 1]) - u(p[:i])
    values /= math.factorial(n)

    return ValuationResult(
        algorithm="permutation_exact_shapley",
        status=Status.Converged,
        values=values,
        data_names=u.data.data_names,
    )


def _combinatorial_exact_shapley(
    indices: NDArray[np.int_], u: Utility, progress: bool
) -> NDArray:
    """Helper function for
    [combinatorial_exact_shapley()][pydvl.value.shapley.naive.combinatorial_exact_shapley].

    Computes the marginal utilities for the set of indices passed and returns
    the value of the samples according to the exact combinatorial definition.
    """
    n = len(u.data)
    local_values = np.zeros(n)
    for i in indices:
        subset: NDArray[np.int_] = np.setxor1d(
            u.data.indices, [i], assume_unique=True
        ).astype(np.int_)
        for s in tqdm(  # type: ignore
            powerset(subset),
            disable=not progress,
            desc=f"Index {i}",
            total=2 ** (n - 1),
            position=0,
        ):
            local_values[i] += (u({i}.union(s)) - u(s)) / math.comb(n - 1, len(s))  # type: ignore
    return local_values / n


def combinatorial_exact_shapley(
    u: Utility,
    *,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
) -> ValuationResult:
    r"""Computes the exact Shapley value using the combinatorial definition.

    $$v_u(i) = \frac{1}{n} \sum_{S \subseteq N \setminus \{i\}} \binom{n-1}{ | S | }^{-1} [u(S \cup \{i\}) − u(S)].$$

    See [Data valuation][computing-data-values] for details.

    !!! Note
        If the length of the training set is > n_jobs*20 this prints a warning
        because the computation is very expensive. Used mostly for internal testing
        and simple use cases. Please refer to the
        [Monte Carlo][pydvl.value.shapley.montecarlo] approximations for practical
        applications.

    Args:
        u: Utility object with model, data, and scoring function
        n_jobs: Number of parallel jobs to use
        config: Object configuring parallel computation, with cluster address,
            number of cpus, etc.
        progress: Whether to display progress bars for each job.

    Returns:
        Object with the data values.
    """
    # Arbitrary choice, will depend on time required, caching, etc.
    if len(u.data) // n_jobs > 20:
        warnings.warn(
            f"Large dataset! Computation requires 2^{len(u.data)} calls to model.fit()"
        )

    def reduce_fun(results: List[NDArray]) -> NDArray:
        return np.array(results).sum(axis=0)  # type: ignore

    map_reduce_job: MapReduceJob[NDArray, NDArray] = MapReduceJob(
        u.data.indices,
        map_func=_combinatorial_exact_shapley,
        map_kwargs=dict(u=u, progress=progress),
        reduce_func=reduce_fun,
        n_jobs=n_jobs,
        config=config,
    )
    values = map_reduce_job()
    return ValuationResult(
        algorithm="combinatorial_exact_shapley",
        status=Status.Converged,
        values=values,
        data_names=u.data.data_names,
    )



"""
This module implements several implementations of [InfluenceFunctionModel]
[pydvl.influence.base_influence_function_model.InfluenceFunctionModel]
utilizing PyTorch.
"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Callable, Optional

import torch
from torch import nn as nn
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from pydvl.utils.progress import log_duration

from ..base_influence_function_model import (
    InfluenceFunctionModel,
    InfluenceMode,
    UnsupportedInfluenceModeException,
)
from .functional import (
    LowRankProductRepresentation,
    create_batch_hvp_function,
    create_hvp_function,
    create_matrix_jacobian_product_function,
    create_per_sample_gradient_function,
    create_per_sample_mixed_derivative_function,
    hessian,
    model_hessian_low_rank,
)
from .util import flatten_dimensions

logger = logging.getLogger(__name__)


class TorchInfluenceFunctionModel(
    InfluenceFunctionModel[torch.Tensor, DataLoader], ABC
):
    """
    Abstract base class for influence computation related to torch models
    """

    def __init__(
        self,
        model: nn.Module,
        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    ):
        self.loss = loss
        self.model = model
        self._n_parameters = sum(
            [p.numel() for p in model.parameters() if p.requires_grad]
        )
        self._model_device = next(
            (p.device for p in model.parameters() if p.requires_grad)
        )
        self._model_params = {
            k: p.detach() for k, p in self.model.named_parameters() if p.requires_grad
        }
        super().__init__()

    @property
    def n_parameters(self):
        return self._n_parameters

    @property
    def is_thread_safe(self) -> bool:
        return False

    @property
    def model_device(self):
        return self._model_device

    @property
    def model_params(self):
        return self._model_params

    @log_duration
    def _loss_grad(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        grads = create_per_sample_gradient_function(self.model, self.loss)(
            self.model_params, x, y
        )
        shape = (x.shape[0], -1)
        return flatten_dimensions(grads.values(), shape=shape)

    @log_duration
    def _flat_loss_mixed_grad(self, x: torch.Tensor, y: torch.Tensor):
        mixed_grads = create_per_sample_mixed_derivative_function(
            self.model, self.loss
        )(self.model_params, x, y)
        shape = (*x.shape, -1)
        return flatten_dimensions(mixed_grads.values(), shape=shape)

    def influences(
        self,
        x_test: torch.Tensor,
        y_test: torch.Tensor,
        x: Optional[torch.Tensor] = None,
        y: Optional[torch.Tensor] = None,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> torch.Tensor:
        r"""
        Compute the approximation of

        \[
        \langle H^{-1}\nabla_{theta} \ell(y_{\text{test}}, f_{\theta}(x_{\text{test}})),
            \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle
        \]

        for the case of up-weighting influence, resp.

        \[
        \langle H^{-1}\nabla_{theta} \ell(y_{\text{test}}, f_{\theta}(x_{\text{test}})),
            \nabla_{x} \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle
        \]

        for the perturbation type influence case.

        Args:
            x_test: model input to use in the gradient computations
                of $H^{-1}\nabla_{theta} \ell(y_{\text{test}},
                    f_{\theta}(x_{\text{test}}))$
            y_test: label tensor to compute gradients
            x: optional model input to use in the gradient computations
                $\nabla_{theta}\ell(y, f_{\theta}(x))$,
                resp. $\nabla_{x}\nabla_{theta}\ell(y, f_{\theta}(x))$,
                if None, use $x=x_{\text{test}}$
            y: optional label tensor to compute gradients
            mode: enum value of [InfluenceType]
                [pydvl.influence.base_influence_model.InfluenceType]

        Returns:
            Tensor representing the element-wise scalar products for the provided batch

        """
        t: torch.Tensor = super().influences(x_test, y_test, x, y, mode=mode)
        return t

    def _influences(
        self,
        x_test: torch.Tensor,
        y_test: torch.Tensor,
        x: Optional[torch.Tensor] = None,
        y: Optional[torch.Tensor] = None,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> torch.Tensor:

        if not self.is_fitted:
            raise ValueError(
                "Instance must be fitted before calling influence methods on it"
            )

        if x is None:

            if y is not None:
                raise ValueError(
                    "Providing labels y, without providing model input x "
                    "is not supported"
                )

            return self._symmetric_values(
                x_test.to(self.model_device),
                y_test.to(self.model_device),
                mode,
            )

        if y is None:
            raise ValueError(
                "Providing model input x without providing labels y is not supported"
            )

        return self._non_symmetric_values(
            x_test.to(self.model_device),
            y_test.to(self.model_device),
            x.to(self.model_device),
            y.to(self.model_device),
            mode,
        )

    def _non_symmetric_values(
        self,
        x_test: torch.Tensor,
        y_test: torch.Tensor,
        x: torch.Tensor,
        y: torch.Tensor,
        mode: InfluenceMode = InfluenceMode.Up,
    ):
        if mode == InfluenceMode.Up:
            if x_test.shape[0] <= x.shape[0]:
                factor = self.influence_factors(x_test, y_test)
                values = self.influences_from_factors(factor, x, y, mode=mode)
            else:
                factor = self.influence_factors(x, y)
                values = self.influences_from_factors(
                    factor, x_test, y_test, mode=mode
                ).T
        elif mode == InfluenceMode.Perturbation:
            factor = self.influence_factors(x_test, y_test)
            values = self.influences_from_factors(factor, x, y, mode=mode)
        else:
            raise UnsupportedInfluenceModeException(mode)
        return values

    def _symmetric_values(
        self, x: torch.Tensor, y: torch.Tensor, mode: InfluenceMode
    ) -> torch.Tensor:

        grad = self._loss_grad(x, y)
        fac = self._solve_hvp(grad)

        if mode == InfluenceMode.Up:
            values = fac @ grad.T
        elif mode == InfluenceMode.Perturbation:
            values = self.influences_from_factors(fac, x, y, mode=mode)
        else:
            raise UnsupportedInfluenceModeException(mode)
        return values

    def influence_factors(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        r"""
        Compute approximation of

        \[ H^{-1}\nabla_{\theta} \ell(y, f_{\theta}(x)) \]

        where the gradient is meant to be per sample of the batch $(x, y)$.

        Args:
            x: model input to use in the gradient computations
            y: label tensor to compute gradients

        Returns:
            Tensor representing the element-wise inverse Hessian matrix vector products

        """
        return super().influence_factors(x, y)

    def _influence_factors(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:

        if not self.is_fitted:
            raise ValueError(
                "Instance must be fitted before calling influence methods on it"
            )

        return self._solve_hvp(
            self._loss_grad(x.to(self.model_device), y.to(self.model_device))
        )

    def influences_from_factors(
        self,
        z_test_factors: torch.Tensor,
        x: torch.Tensor,
        y: torch.Tensor,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> torch.Tensor:
        r"""
        Computation of

        \[ \langle z_{\text{test_factors}},
            \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the case of up-weighting influence, resp.

        \[ \langle z_{\text{test_factors}},
            \nabla_{x} \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the perturbation type influence case. The gradient is meant to be per sample
        of the batch $(x, y)$.

        Args:
             z_test_factors: pre-computed tensor, approximating
                $H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
                    f_{\theta}(x_{\text{test}}))$
             x: model input to use in the gradient computations
                $\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                resp. $\nabla_{x}\nabla_{\theta}\ell(y, f_{\theta}(x))$
             y: label tensor to compute gradients
             mode: enum value of [InfluenceType]
                [pydvl.influence.twice_differentiable.InfluenceType]

        Returns:
            Tensor representing the element-wise scalar products for the provided batch

        """
        if mode == InfluenceMode.Up:
            return (
                z_test_factors
                @ self._loss_grad(x.to(self.model_device), y.to(self.model_device)).T
            )
        elif mode == InfluenceMode.Perturbation:
            return torch.einsum(
                "ia,j...a->ij...",
                z_test_factors,
                self._flat_loss_mixed_grad(
                    x.to(self.model_device), y.to(self.model_device)
                ),
            )
        else:
            raise UnsupportedInfluenceModeException(mode)

    @abstractmethod
    def _solve_hvp(self, rhs: torch.Tensor) -> torch.Tensor:
        pass


class DirectInfluence(TorchInfluenceFunctionModel):
    r"""
    Given a model and training data, it finds x such that \(Hx = b\),
    with \(H\) being the model hessian.

    Args:
        model: instance of [torch.nn.Module][torch.nn.Module].
        hessian_regularization: Regularization of the hessian.
    """

    def __init__(
        self,
        model: nn.Module,
        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        hessian_regularization: float = 0.0,
    ):
        super().__init__(model, loss)
        self.hessian_regularization = hessian_regularization

    hessian: torch.Tensor

    @property
    def is_fitted(self):
        try:
            return self.hessian is not None
        except AttributeError:
            return False

    def fit(self, data: DataLoader) -> DirectInfluence:
        """
        Compute the hessian matrix based on a provided dataloader

        Args:
            data: Instance of [torch.utils.data.Dataloader]
                [torch.utils.data.Dataloader]

        Returns:
            The fitted instance
        """
        self.hessian = hessian(self.model, self.loss, data)
        return self

    @log_duration
    def influences(
        self,
        x_test: torch.Tensor,
        y_test: torch.Tensor,
        x: Optional[torch.Tensor] = None,
        y: Optional[torch.Tensor] = None,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> torch.Tensor:
        r"""
        Compute approximation of

        \[ \langle H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
            f_{\theta}(x_{\text{test}})),
            \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle, \]

        for the case of up-weighting influence, resp.

        \[ \langle H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
            f_{\theta}(x_{\text{test}})),
            \nabla_{x} \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the perturbation type influence case. The action of $H^{-1}$ is achieved
        via a direct solver using [torch.linalg.solve][torch.linalg.solve].

        Args:
            x_test: model input to use in the gradient computations of
                $H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
                    f_{\theta}(x_{\text{test}}))$
            y_test: label tensor to compute gradients
            x: optional model input to use in the gradient computations
                $\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                resp. $\nabla_{x}\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                if None, use $x=x_{\text{test}}$
            y: optional label tensor to compute gradients
            mode: enum value of [InfluenceType]
                [pydvl.influence.base_influence_model.InfluenceType]

        Returns:
            [torch.nn.Tensor][torch.nn.Tensor] representing the element-wise
                scalar products for the provided batch.

        """
        return super().influences(x_test, y_test, x, y, mode=mode)

    @log_duration
    def _solve_hvp(self, rhs: torch.Tensor) -> torch.Tensor:
        return torch.linalg.solve(
            self.hessian.to(self.model_device)
            + self.hessian_regularization
            * torch.eye(self.n_parameters, device=self.model_device),
            rhs.T.to(self.model_device),
        ).T

    def to(self, device: torch.device):
        self.hessian = self.hessian.to(device)
        self.model = self.model.to(device)
        self._model_device = device
        self._model_params = {
            k: p.detach().to(device)
            for k, p in self.model.named_parameters()
            if p.requires_grad
        }
        return self


class CgInfluence(TorchInfluenceFunctionModel):
    r"""
    Given a model and training data, it uses conjugate gradient to calculate the
    inverse of the Hessian Vector Product. More precisely, it finds x such that \(Hx =
    b\), with \(H\) being the model hessian. For more info, see
    [Conjugate Gradient][conjugate-gradient].

    Args:
        model: Instance of [torch.nn.Module][torch.nn.Module].
        loss: A callable that takes the model's output and target as input and returns
              the scalar loss.
        hessian_regularization: Regularization of the hessian.
        x0: Initial guess for hvp. If None, defaults to b.
        rtol: Maximum relative tolerance of result.
        atol: Absolute tolerance of result.
        maxiter: Maximum number of iterations. If None, defaults to 10*len(b).
        progress: If True, display progress bars.

    """

    def __init__(
        self,
        model: nn.Module,
        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        hessian_regularization: float = 0.0,
        x0: Optional[torch.Tensor] = None,
        rtol: float = 1e-7,
        atol: float = 1e-7,
        maxiter: Optional[int] = None,
        progress: bool = False,
    ):
        super().__init__(model, loss)
        self.progress = progress
        self.maxiter = maxiter
        self.atol = atol
        self.rtol = rtol
        self.x0 = x0
        self.hessian_regularization = hessian_regularization

    train_dataloader: DataLoader

    @property
    def is_fitted(self):
        try:
            return self.train_dataloader is not None
        except AttributeError:
            return False

    def fit(self, data: DataLoader) -> CgInfluence:
        self.train_dataloader = data
        return self

    @log_duration
    def influences(
        self,
        x_test: torch.Tensor,
        y_test: torch.Tensor,
        x: Optional[torch.Tensor] = None,
        y: Optional[torch.Tensor] = None,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> torch.Tensor:
        r"""
        Compute approximation of

        \[ \langle H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
            f_{\theta}(x_{\text{test}})),
            \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle, \]

        for the case of up-weighting influence, resp.

        \[ \langle H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
            f_{\theta}(x_{\text{test}})),
            \nabla_{x} \nabla_{\theta} \ell(y, f_{\theta}(x)) \rangle \]

        for the perturbation type influence case. The approximate action of $H^{-1}$
        is achieved via the [conjugate gradient method]
        (https://en.wikipedia.org/wiki/Conjugate_gradient_method).

        Args:
            x_test: model input to use in the gradient computations of
                $H^{-1}\nabla_{\theta} \ell(y_{\text{test}},
                    f_{\theta}(x_{\text{test}}))$
            y_test: label tensor to compute gradients
            x: optional model input to use in the gradient computations
                $\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                resp. $\nabla_{x}\nabla_{\theta}\ell(y, f_{\theta}(x))$,
                if None, use $x=x_{\text{test}}$
            y: optional label tensor to compute gradients
            mode: enum value of [InfluenceType]
                [pydvl.influence.base_influence_model.InfluenceType]

        Returns:
            [torch.nn.Tensor][torch.nn.Tensor] representing the element-wise
                scalar products for the provided batch.

        """
        return super().influences(x_test, y_test, x, y, mode=mode)

    @log_duration
    def _solve_hvp(self, rhs: torch.Tensor) -> torch.Tensor:
        if len(self.train_dataloader) == 0:
            raise ValueError("Training dataloader must not be empty.")

        hvp = create_hvp_function(self.model, self.loss, self.train_dataloader)

        def reg_hvp(v: torch.Tensor):
            return hvp(v) + self.hessian_regularization * v.type(rhs.dtype)

        batch_cg = torch.zeros_like(rhs)

        for idx, bi in enumerate(
            tqdm(rhs, disable=not self.progress, desc="Conjugate gradient")
        ):
            batch_result = self._solve_cg(
                reg_hvp,
                bi,
                x0=self.x0,
                rtol=self.rtol,
                atol=self.atol,
                maxiter=self.maxiter,
            )
            batch_cg[idx] = batch_result
        return batch_cg

    def to(self, device: torch.device):
        self.model = self.model.to(device)
        self._model_params = {
            k: p.detach().to(device)
            for k, p in self.model.named_parameters()
            if p.requires_grad
        }
        self._model_device = device
        return self

    @staticmethod
    def _solve_cg(
        hvp: Callable[[torch.Tensor], torch.Tensor],
        b: torch.Tensor,
        *,
        x0: Optional[torch.Tensor] = None,
        rtol: float = 1e-7,
        atol: float = 1e-7,
        maxiter: Optional[int] = None,
    ) -> torch.Tensor:
        r"""
        Conjugate gradient solver for the Hessian vector product.

        Args:
            hvp: A callable Hvp, operating with tensors of size N.
            b: A vector or matrix, the right hand side of the equation \(Hx = b\).
            x0: Initial guess for hvp.
            rtol: Maximum relative tolerance of result.
            atol: Absolute tolerance of result.
            maxiter: Maximum number of iterations. If None, defaults to 10*len(b).

        Returns:
            [torch.nn.Tensor][torch.nn.Tensor] representing the solution of \(Ax=b\).
        """

        if x0 is None:
            x0 = torch.clone(b)
        if maxiter is None:
            maxiter = len(b) * 10

        y_norm = torch.sum(torch.matmul(b, b)).item()
        stopping_val = max([rtol**2 * y_norm, atol**2])

        x = x0
        p = r = (b - hvp(x)).squeeze()
        gamma = torch.sum(torch.matmul(r, r)).item()

        for k in range(maxiter):
            if gamma < stopping_val:
                break
            Ap = hvp(p).squeeze()
            alpha = gamma / torch.sum(torch.matmul(p, Ap)).item()
            x += alpha * p
            r -= alpha * Ap
            gamma_ = torch.sum(torch.matmul(r, r)).item()
            beta = gamma_ / gamma
            gamma = gamma_
            p = r + beta * p

        return x


class LissaInfluence(TorchInfluenceFunctionModel):
    r"""
    Uses LISSA, Linear time Stochastic Second-Order Algorithm, to iteratively
    approximate the inverse Hessian. More precisely, it finds x s.t. \(Hx = b\),
    with \(H\) being the model's second derivative wrt. the parameters.
    This is done with the update

    \[H^{-1}_{j+1} b = b + (I - d) \ H - \frac{H^{-1}_j b}{s},\]

    where \(I\) is the identity matrix, \(d\) is a dampening term and \(s\) a scaling
    factor that are applied to help convergence. For details,
    see [Linear time Stochastic Second-Order Approximation (LiSSA)]
    [linear-time-stochastic-second-order-approximation-lissa]

    Args:
        model: instance of [torch.nn.Module][torch.nn.Module].
        hessian_regularization: Regularization of the hessian.
        maxiter: Maximum number of iterations.
        dampen: Dampening factor, defaults to 0 for no dampening.
        scale: Scaling factor, defaults to 10.
        h0: Initial guess for hvp.
        rtol: tolerance to use for early stopping
        progress: If True, display progress bars.
    """

    def __init__(
        self,
        model: nn.Module,
        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        hessian_regularization: float = 0.0,
        maxiter: int = 1000,
        dampen: float = 0.0,
        scale: float = 10.0,
        h0: Optional[torch.Tensor] = None,
        rtol: float = 1e-4,
        progress: bool = False,
    ):
        super().__init__(model, loss)
        self.maxiter = maxiter
        self.hessian_regularization = hessian_regularization
        self.progress = progress
        self.rtol = rtol
        self.h0 = h0
        self.scale = scale
        self.dampen = dampen

    train_dataloader: DataLoader

    @property
    def is_fitted(self):
        try:
            return self.train_dataloader is not None
        except AttributeError:
            return False

    def fit(self, data: DataLoader) -> LissaInfluence:
        self.train_dataloader = data
        return self

    @log_duration
    def _solve_hvp(self, rhs: torch.Tensor) -> torch.Tensor:

        h_estimate = self.h0 if self.h0 is not None else torch.clone(rhs)

        shuffled_training_data = DataLoader(
            self.train_dataloader.dataset,
            self.train_dataloader.batch_size,
            shuffle=True,
        )

        def lissa_step(
            h: torch.Tensor, reg_hvp: Callable[[torch.Tensor], torch.Tensor]
        ) -> torch.Tensor:
            """Given an estimate of the hessian inverse and the regularised hessian
            vector product, it computes the next estimate.

            Args:
                h: An estimate of the hessian inverse.
                reg_hvp: Regularised hessian vector product.

            Returns:
                The next estimate of the hessian inverse.
            """
            return rhs + (1 - self.dampen) * h - reg_hvp(h) / self.scale

        model_params = {
            k: p.detach() for k, p in self.model.named_parameters() if p.requires_grad
        }
        b_hvp = torch.vmap(
            create_batch_hvp_function(self.model, self.loss),
            in_dims=(None, None, None, 0),
        )
        for _ in tqdm(range(self.maxiter), disable=not self.progress, desc="Lissa"):
            x, y = next(iter(shuffled_training_data))
            # grad_xy = model.grad(x, y, create_graph=True)
            reg_hvp = (
                lambda v: b_hvp(model_params, x, y, v) + self.hessian_regularization * v
            )
            residual = lissa_step(h_estimate, reg_hvp) - h_estimate
            h_estimate += residual
            if torch.isnan(h_estimate).any():
                raise RuntimeError("NaNs in h_estimate. Increase scale or dampening.")
            max_residual = torch.max(torch.abs(residual / h_estimate))
            if max_residual < self.rtol:
                break

        mean_residual = torch.mean(torch.abs(residual / h_estimate))

        logger.info(
            f"Terminated Lissa with {max_residual*100:.2f} % max residual."
            f" Mean residual: {mean_residual*100:.5f} %"
        )
        return h_estimate / self.scale


class ArnoldiInfluence(TorchInfluenceFunctionModel):
    r"""
    Solves the linear system Hx = b, where H is the Hessian of the model's loss function
    and b is the given right-hand side vector.
    It employs the [implicitly restarted Arnoldi method]
    (https://en.wikipedia.org/wiki/Arnoldi_iteration) for
    computing a partial eigen decomposition, which is used fo the inversion i.e.

    \[x = V D^{-1} V^T b\]

    where \(D\) is a diagonal matrix with the top (in absolute value) `rank_estimate`
    eigenvalues of the Hessian
    and \(V\) contains the corresponding eigenvectors.
    For more information, see [Arnoldi][arnoldi].

    Args:
        model: Instance of [torch.nn.Module][torch.nn.Module].
            The Hessian will be calculated with respect to this model's parameters.
        hessian_regularization: Optional regularization parameter added
            to the Hessian-vector product for numerical stability.
        rank_estimate: The number of eigenvalues and corresponding eigenvectors
            to compute. Represents the desired rank of the Hessian approximation.
        krylov_dimension: The number of Krylov vectors to use for the Lanczos method.
            Defaults to min(model's number of parameters,
            max(2 times rank_estimate + 1, 20)).
        tol: The stopping criteria for the Lanczos algorithm.
            Ignored if `low_rank_representation` is provided.
        max_iter: The maximum number of iterations for the Lanczos method.
            Ignored if `low_rank_representation` is provided.
        eigen_computation_on_gpu: If True, tries to execute the eigen pair approximation
            on the model's device
            via a cupy implementation. Ensure the model size or rank_estimate
            is appropriate for device memory.
            If False, the eigen pair approximation is executed on the CPU by the scipy
            wrapper to ARPACK.
    """
    low_rank_representation: LowRankProductRepresentation

    def __init__(
        self,
        model,
        loss,
        hessian_regularization: float = 0.0,
        rank_estimate: int = 10,
        krylov_dimension: Optional[int] = None,
        tol: float = 1e-6,
        max_iter: Optional[int] = None,
        eigen_computation_on_gpu: bool = False,
    ):

        super().__init__(model, loss)
        self.hessian_regularization = hessian_regularization
        self.rank_estimate = rank_estimate
        self.tol = tol
        self.max_iter = max_iter
        self.krylov_dimension = krylov_dimension
        self.eigen_computation_on_gpu = eigen_computation_on_gpu

    @property
    def is_fitted(self):
        try:
            return self.low_rank_representation is not None
        except AttributeError:
            return False

    def fit(self, data: DataLoader) -> ArnoldiInfluence:
        r"""
        Fitting corresponds to the computation of the low rank decomposition

        \[ V D^{-1} V^T \]

        of the Hessian defined by the provided data loader.

        Args:
            data: Instance of [torch.utils.data.Dataloader][torch.utils.data.Dataloader]

        Returns:
            The fitted instance

        """
        low_rank_representation = model_hessian_low_rank(
            self.model,
            self.loss,
            data,
            hessian_perturbation=0.0,  # regularization is applied, when computing values
            rank_estimate=self.rank_estimate,
            krylov_dimension=self.krylov_dimension,
            tol=self.tol,
            max_iter=self.max_iter,
            eigen_computation_on_gpu=self.eigen_computation_on_gpu,
        )
        self.low_rank_representation = low_rank_representation.to(self.model_device)
        return self

    def _non_symmetric_values(
        self,
        x_test: torch.Tensor,
        y_test: torch.Tensor,
        x: torch.Tensor,
        y: torch.Tensor,
        mode: InfluenceMode = InfluenceMode.Up,
    ) -> torch.Tensor:

        if mode == InfluenceMode.Up:
            mjp = create_matrix_jacobian_product_function(
                self.model, self.loss, self.low_rank_representation.projections.T
            )
            left = mjp(self.model_params, x_test, y_test)

            regularized_eigenvalues = (
                self.low_rank_representation.eigen_vals + self.hessian_regularization
            )

            right = torch.diag_embed(1.0 / regularized_eigenvalues) @ mjp(
                self.model_params, x, y
            )
            values = torch.einsum("ij, ik -> jk", left, right)
        elif mode == InfluenceMode.Perturbation:
            factors = self.influence_factors(x_test, y_test)
            values = self.influences_from_factors(factors, x, y, mode=mode)
        else:
            raise UnsupportedInfluenceModeException(mode)
        return values

    def _symmetric_values(
        self, x: torch.Tensor, y: torch.Tensor, mode: InfluenceMode
    ) -> torch.Tensor:

        if mode == InfluenceMode.Up:
            left = create_matrix_jacobian_product_function(
                self.model, self.loss, self.low_rank_representation.projections.T
            )(self.model_params, x, y)
            regularized_eigenvalues = (
                self.low_rank_representation.eigen_vals + self.hessian_regularization
            )
            right = torch.diag_embed(1.0 / regularized_eigenvalues) @ left
            values = torch.einsum("ij, ik -> jk", left, right)
        elif mode == InfluenceMode.Perturbation:
            factors = self.influence_factors(x, y)
            values = self.influences_from_factors(factors, x, y, mode=mode)
        else:
            raise UnsupportedInfluenceModeException(mode)
        return values

    @log_duration
    def _solve_hvp(self, rhs: torch.Tensor) -> torch.Tensor:

        regularized_eigenvalues = (
            self.low_rank_representation.eigen_vals + self.hessian_regularization
        )

        result = self.low_rank_representation.projections @ (
            torch.diag_embed(1.0 / regularized_eigenvalues)
            @ (self.low_rank_representation.projections.t() @ rhs.t())
        )

        return result.t()

    def to(self, device: torch.device):
        return ArnoldiInfluence(
            self.model.to(device), self.loss, self.low_rank_representation.to(device)
        )



r"""
Class-wise Shapley (Schoch et al., 2022)[^1] offers a Shapley framework tailored
for classification problems. Let $D$ be a dataset, $D_{y_i}$ be the subset of
$D$ with labels $y_i$, and $D_{-y_i}$ be the complement of $D_{y_i}$ in $D$. The
key idea is that a sample $(x_i, y_i)$, might enhance the overall performance on
$D$, while being detrimental for the performance on $D_{y_i}$. The Class-wise
value is defined as:

$$
v_u(i) = \frac{1}{2^{|D_{-y_i}|}} \sum_{S_{-y_i}} \frac{1}{|D_{y_i}|!}
\sum_{S_{y_i}} \binom{|D_{y_i}|-1}{|S_{y_i}|}^{-1}
[u( S_{y_i} \cup \{i\} | S_{-y_i} ) − u( S_{y_i} | S_{-y_i})],
$$

where $S_{y_i} \subseteq D_{y_i} \setminus \{i\}$ and $S_{-y_i} \subseteq
D_{-y_i}$.

!!! tip "Analysis of Class-wise Shapley"
    For a detailed analysis of the method, with comparison to other valuation
    techniques, please refer to the [main
    documentation](../../../../../value/classwise-shapley).

In practice, the quantity above is estimated using Monte Carlo sampling of
the powerset and the set of index permutations. This results in the estimator

$$
v_u(i) = \frac{1}{K} \sum_k \frac{1}{L} \sum_l
[u(\sigma^{(l)}_{:i} \cup \{i\} | S^{(k)} ) − u( \sigma^{(l)}_{:i} | S^{(k)})],
$$

with $S^{(1)}, \dots, S^{(K)} \subseteq T_{-y_i},$ $\sigma^{(1)}, \dots,
\sigma^{(L)} \in \Pi(T_{y_i}\setminus\{i\}),$ and $\sigma^{(l)}_{:i}$ denoting
the set of indices in permutation $\sigma^{(l)}$ before the position where $i$
appears. The sets $T_{y_i}$ and $T_{-y_i}$ are the training sets for the labels
$y_i$ and $-y_i$, respectively.

??? info "Notes for derivation of test cases"
    The unit tests include the following manually constructed data:
    Let $D=\{(1,0),(2,0),(3,0),(4,1)\}$ be the test set and $T=\{(1,0),(2,0),(3,1),(4,1)\}$
    the train set. This specific dataset is chosen as it allows to solve the model

    $$y = \max(0, \min(1, \text{round}(\beta^T x)))$$

    in closed form $\beta = \frac{\text{dot}(x, y)}{\text{dot}(x, x)}$. From the closed-form
    solution, the tables for in-class accuracy $a_S(D_{y_i})$ and out-of-class accuracy 
    $a_S(D_{-y_i})$ can be calculated. By using these tables and setting 
    $\{S^{(1)}, \dots, S^{(K)}\} = 2^{T_{-y_i}}$ and 
    $\{\sigma^{(1)}, \dots, \sigma^{(L)}\} = \Pi(T_{y_i}\setminus\{i\})$,
    the Monte Carlo estimator can be evaluated ($2^M$ is the powerset of $M$).
    The details of the derivation are left to the eager reader.

# References

[^1]: <a name="schoch_csshapley_2022"></a>Schoch, Stephanie, Haifeng Xu, and
    Yangfeng Ji. [CS-Shapley: Class-wise Shapley Values for Data Valuation in
    Classification](https://openreview.net/forum?id=KTOcrOR5mQ9). In Proc. of
    the Thirty-Sixth Conference on Neural Information Processing Systems
    (NeurIPS). New Orleans, Louisiana, USA, 2022.

"""
import logging
import numbers
from concurrent.futures import FIRST_COMPLETED, Future, wait
from copy import copy
from typing import Callable, Optional, Set, Tuple, Union, cast

import numpy as np
from numpy.random import SeedSequence
from numpy.typing import NDArray
from tqdm import tqdm

from pydvl.parallel import (
    ParallelConfig,
    effective_n_jobs,
    init_executor,
    init_parallel_backend,
)
from pydvl.utils import (
    Dataset,
    Scorer,
    ScorerCallable,
    Seed,
    SupervisedModel,
    Utility,
    ensure_seed_sequence,
    random_powerset_label_min,
)
from pydvl.value.result import ValuationResult
from pydvl.value.shapley.truncated import TruncationPolicy
from pydvl.value.stopping import MaxChecks, StoppingCriterion

logger = logging.getLogger(__name__)

__all__ = ["ClasswiseScorer", "compute_classwise_shapley_values"]


class ClasswiseScorer(Scorer):
    r"""A Scorer designed for evaluation in classification problems. Its value
    is computed from an in-class and an out-of-class "inner score" (Schoch et
    al., 2022) <sup><a href="#schoch_csshapley_2022">1</a></sup>. Let $S$ be the
    training set and $D$ be the valuation set. For each label $c$, $D$ is
    factorized into two disjoint sets: $D_c$ for in-class instances and $D_{-c}$
    for out-of-class instances. The score combines an in-class metric of
    performance, adjusted by a discounted out-of-class metric. These inner
    scores must be provided upon construction or default to accuracy. They are
    combined into:

    $$
    u(S_{y_i}) = f(a_S(D_{y_i}))\ g(a_S(D_{-y_i})),
    $$

    where $f$ and $g$ are continuous, monotonic functions. For a detailed
    explanation, refer to section four of (Schoch et al., 2022)<sup><a
    href="#schoch_csshapley_2022"> 1</a></sup>.

    !!! warning Multi-class support
        Metrics must support multiple class labels if you intend to apply them
        to a multi-class problem. For instance, the metric 'accuracy' supports
        multiple classes, but the metric `f1` does not. For a two-class
        classification problem, using `f1_weighted` is essentially equivalent to
        using `accuracy`.

    Args:
        scoring: Name of the scoring function or a callable that can be passed
            to [Scorer][pydvl.utils.score.Scorer].
        default: Score to use when a model fails to provide a number, e.g. when
            too little was used to train it, or errors arise.
        range: Numerical range of the score function. Some Monte Carlo methods
            can use this to estimate the number of samples required for a
            certain quality of approximation. If not provided, it can be read
            from the `scoring` object if it provides it, for instance if it was
            constructed with
            [compose_score][pydvl.utils.score.compose_score].
        in_class_discount_fn: Continuous, monotonic increasing function used to
            discount the in-class score.
        out_of_class_discount_fn: Continuous, monotonic increasing function used
            to discount the out-of-class score.
        initial_label: Set initial label (for the first iteration)
        name: Name of the scorer. If not provided, the name of the inner scoring
            function will be prefixed by `classwise `.

    !!! tip "New in version 0.7.1"
    """

    def __init__(
        self,
        scoring: Union[str, ScorerCallable] = "accuracy",
        default: float = 0.0,
        range: Tuple[float, float] = (0, 1),
        in_class_discount_fn: Callable[[float], float] = lambda x: x,
        out_of_class_discount_fn: Callable[[float], float] = np.exp,
        initial_label: Optional[int] = None,
        name: Optional[str] = None,
    ):
        disc_score_in_class = in_class_discount_fn(range[1])
        disc_score_out_of_class = out_of_class_discount_fn(range[1])
        transformed_range = (0, disc_score_in_class * disc_score_out_of_class)
        super().__init__(
            scoring=scoring,
            range=transformed_range,
            default=default,
            name=name or f"classwise {str(scoring)}",
        )
        self._in_class_discount_fn = in_class_discount_fn
        self._out_of_class_discount_fn = out_of_class_discount_fn
        self.label = initial_label

    def __str__(self):
        return self._name

    def __call__(
        self: "ClasswiseScorer",
        model: SupervisedModel,
        x_test: NDArray[np.float_],
        y_test: NDArray[np.int_],
    ) -> float:
        (
            in_class_score,
            out_of_class_score,
        ) = self.estimate_in_class_and_out_of_class_score(model, x_test, y_test)
        disc_score_in_class = self._in_class_discount_fn(in_class_score)
        disc_score_out_of_class = self._out_of_class_discount_fn(out_of_class_score)
        return disc_score_in_class * disc_score_out_of_class

    def estimate_in_class_and_out_of_class_score(
        self,
        model: SupervisedModel,
        x_test: NDArray[np.float_],
        y_test: NDArray[np.int_],
        rescale_scores: bool = True,
    ) -> Tuple[float, float]:
        r"""
        Computes in-class and out-of-class scores using the provided inner
        scoring function. The result is

        $$
        a_S(D=\{(x_1, y_1), \dots, (x_K, y_K)\}) = \frac{1}{N} \sum_k s(y(x_k), y_k).
        $$

        In this context, for label $c$ calculations are executed twice: once for $D_c$
        and once for $D_{-c}$ to determine the in-class and out-of-class scores,
        respectively. By default, the raw scores are multiplied by $\frac{|D_c|}{|D|}$
        and $\frac{|D_{-c}|}{|D|}$, respectively. This is done to ensure that both
        scores are of the same order of magnitude. This normalization is particularly
        useful when the inner score function $a_S$ is calculated by an estimator of the
        form $\frac{1}{N} \sum_i x_i$, e.g. the accuracy.

        Args:
            model: Model used for computing the score on the validation set.
            x_test: Array containing the features of the classification problem.
            y_test: Array containing the labels of the classification problem.
            rescale_scores: If set to True, the scores will be denormalized. This is
                particularly useful when the inner score function $a_S$ is calculated by
                an estimator of the form $\frac{1}{N} \sum_i x_i$.

        Returns:
            Tuple containing the in-class and out-of-class scores.
        """
        scorer = self._scorer
        label_set_match = y_test == self.label
        label_set = np.where(label_set_match)[0]
        num_classes = len(np.unique(y_test))

        if len(label_set) == 0:
            return 0, 1 / (num_classes - 1)

        complement_label_set = np.where(~label_set_match)[0]
        in_class_score = scorer(model, x_test[label_set], y_test[label_set])
        out_of_class_score = scorer(
            model, x_test[complement_label_set], y_test[complement_label_set]
        )

        if rescale_scores:
            n_in_class = np.count_nonzero(y_test == self.label)
            n_out_of_class = len(y_test) - n_in_class
            in_class_score *= n_in_class / (n_in_class + n_out_of_class)
            out_of_class_score *= n_out_of_class / (n_in_class + n_out_of_class)

        return in_class_score, out_of_class_score


def compute_classwise_shapley_values(
    u: Utility,
    *,
    done: StoppingCriterion,
    truncation: TruncationPolicy,
    done_sample_complements: Optional[StoppingCriterion] = None,
    normalize_values: bool = True,
    use_default_scorer_value: bool = True,
    min_elements_per_label: int = 1,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
    seed: Optional[Seed] = None,
) -> ValuationResult:
    r"""
    Computes an approximate Class-wise Shapley value by sampling independent
    permutations of the index set for each label and index sets sampled from the
    powerset of the complement (with respect to the currently evaluated label),
    approximating the sum:

    $$
    v_u(i) = \frac{1}{K} \sum_k \frac{1}{L} \sum_l
    [u(\sigma^{(l)}_{:i} \cup \{i\} | S^{(k)} ) − u( \sigma^{(l)}_{:i} | S^{(k)})],
    $$

    where $\sigma_{:i}$ denotes the set of indices in permutation sigma before
    the position where $i$ appears and $S$ is a subset of the index set of all other
    labels(see [[data-valuation]] for details).

    Args:
        u: Utility object containing model, data, and scoring function. The
            scorer must be of type
            [ClasswiseScorer][pydvl.value.shapley.classwise.ClasswiseScorer].
        done: Function that checks whether the computation needs to stop.
        truncation: Callable function that decides whether to interrupt processing a
            permutation and set subsequent marginals to zero.
        done_sample_complements: Function checking whether computation needs to stop.
            Otherwise, it will resample conditional sets until the stopping criterion is
            met.
        normalize_values: Indicates whether to normalize the values by the variation
            in each class times their in-class accuracy.
        done_sample_complements: Number of times to resample the complement set
            for each permutation.
        use_default_scorer_value: The first set of indices is the sampled complement
            set. Unless not otherwise specified, the default scorer value is used for
            this. If it is set to false, the base score is calculated from the utility.
        min_elements_per_label: The minimum number of elements for each opposite
            label.
        n_jobs: Number of parallel jobs to run.
        config: Parallel configuration.
        progress: Whether to display a progress bar.
        seed: Either an instance of a numpy random number generator or a seed for it.

    Returns:
        ValuationResult object containing computed data values.

    !!! tip "New in version 0.7.1"
    """
    dim_correct = u.data.y_train.ndim == 1 and u.data.y_test.ndim == 1
    is_integral = all(
        map(
            lambda v: isinstance(v, numbers.Integral), (*u.data.y_train, *u.data.y_test)
        )
    )
    if not dim_correct or not is_integral:
        raise ValueError(
            "The supplied dataset has to be a 1-dimensional classification dataset."
        )

    if not isinstance(u.scorer, ClasswiseScorer):
        raise ValueError(
            "Please set a subclass of ClasswiseScorer object as scorer object of the"
            " utility. See scoring argument of Utility."
        )

    parallel_backend = init_parallel_backend(config)
    u_ref = parallel_backend.put(u)
    n_jobs = effective_n_jobs(n_jobs, config)
    n_submitted_jobs = 2 * n_jobs

    pbar = tqdm(disable=not progress, position=0, total=100, unit="%")
    algorithm = "classwise_shapley"
    accumulated_result = ValuationResult.zeros(
        algorithm=algorithm, indices=u.data.indices, data_names=u.data.data_names
    )
    terminate_exec = False
    seed_sequence = ensure_seed_sequence(seed)

    with init_executor(max_workers=n_jobs, config=config) as executor:
        pending: Set[Future] = set()
        while True:
            completed_futures, pending = wait(
                pending, timeout=60, return_when=FIRST_COMPLETED
            )
            for future in completed_futures:
                accumulated_result += future.result()
                if done(accumulated_result):
                    terminate_exec = True
                    break

            pbar.n = 100 * done.completion()
            pbar.refresh()
            if terminate_exec:
                break

            n_remaining_slots = n_submitted_jobs - len(pending)
            seeds = seed_sequence.spawn(n_remaining_slots)
            for i in range(n_remaining_slots):
                future = executor.submit(
                    _permutation_montecarlo_classwise_shapley_one_step,
                    u_ref,
                    truncation=truncation,
                    done_sample_complements=done_sample_complements,
                    use_default_scorer_value=use_default_scorer_value,
                    min_elements_per_label=min_elements_per_label,
                    algorithm_name=algorithm,
                    seed=seeds[i],
                )
                pending.add(future)

    result = accumulated_result
    if normalize_values:
        result = _normalize_classwise_shapley_values(result, u)

    return result


def _permutation_montecarlo_classwise_shapley_one_step(
    u: Utility,
    *,
    done_sample_complements: Optional[StoppingCriterion] = None,
    truncation: TruncationPolicy,
    use_default_scorer_value: bool = True,
    min_elements_per_label: int = 1,
    algorithm_name: str = "classwise_shapley",
    seed: Optional[SeedSequence] = None,
) -> ValuationResult:
    """Helper function for [compute_classwise_shapley_values()]
    [pydvl.value.shapley.classwise.compute_classwise_shapley_values].


    Args:
        u: Utility object containing model, data, and scoring function. The
            scorer must be of type [ClasswiseScorer]
            [pydvl.value.shapley.classwise.ClasswiseScorer].
        done_sample_complements: Function checking whether computation needs to stop.
            Otherwise, it will resample conditional sets until the stopping criterion is
            met.
        truncation: Callable function that decides whether to interrupt processing a
            permutation and set subsequent marginals to zero.
        use_default_scorer_value: The first set of indices is the sampled complement
            set. Unless not otherwise specified, the default scorer value is used for
            this. If it is set to false, the base score is calculated from the utility.
        min_elements_per_label: The minimum number of elements for each opposite
            label.
        algorithm_name: For the results object.
        seed: Either an instance of a numpy random number generator or a seed for it.

    Returns:
        ValuationResult object containing computed data values.
    """
    if done_sample_complements is None:
        done_sample_complements = MaxChecks(1)

    result = ValuationResult.zeros(
        algorithm=algorithm_name, indices=u.data.indices, data_names=u.data.data_names
    )
    rng = np.random.default_rng(seed)
    x_train, y_train = u.data.get_training_data(u.data.indices)
    unique_labels = np.unique(y_train)
    scorer = cast(ClasswiseScorer, copy(u.scorer))
    u.scorer = scorer

    for label in unique_labels:
        u.scorer.label = label
        class_indices_set, class_complement_indices_set = _split_indices_by_label(
            u.data.indices, y_train, label
        )
        _, complement_y_train = u.data.get_training_data(class_complement_indices_set)
        indices_permutation = rng.permutation(class_indices_set)
        done_sample_complements.reset()

        for subset_idx, subset_complement in enumerate(
            random_powerset_label_min(
                class_complement_indices_set,
                complement_y_train,
                min_elements_per_label=min_elements_per_label,
                seed=rng,
            )
        ):
            result += _permutation_montecarlo_shapley_rollout(
                u,
                indices_permutation,
                additional_indices=subset_complement,
                truncation=truncation,
                algorithm_name=algorithm_name,
                use_default_scorer_value=use_default_scorer_value,
            )
            if done_sample_complements(result):
                break

    return result


def _normalize_classwise_shapley_values(
    result: ValuationResult, u: Utility
) -> ValuationResult:
    r"""
    Normalize a valuation result specific to classwise Shapley.

    Each value $v_i$ associated with the sample $(x_i, y_i)$ is normalized by
    multiplying it with $a_S(D_{y_i})$ and dividing by $\sum_{j \in D_{y_i}} v_j$. For
    more details, see (Schoch et al., 2022) <sup><a href="#schoch_csshapley_2022">1</a>
    </sup>.

    Args:
        result: ValuationResult object to be normalized.
        u: Utility object containing model, data, and scoring function. The
            scorer must be of type [ClasswiseScorer]
            [pydvl.value.shapley.classwise.ClasswiseScorer].

    Returns:
        Normalized ValuationResult object.
    """
    y_train = u.data.y_train
    unique_labels = np.unique(np.concatenate((y_train, u.data.y_test)))
    scorer = cast(ClasswiseScorer, u.scorer)

    for idx_label, label in enumerate(unique_labels):
        scorer.label = label
        active_elements = y_train == label
        indices_label_set = np.where(active_elements)[0]
        indices_label_set = u.data.indices[indices_label_set]

        u.model.fit(u.data.x_train, u.data.y_train)
        scorer.label = label
        in_class_acc, _ = scorer.estimate_in_class_and_out_of_class_score(
            u.model, u.data.x_test, u.data.y_test
        )

        sigma = np.sum(result.values[indices_label_set])
        if sigma != 0:
            result.scale(in_class_acc / sigma, indices=indices_label_set)

    return result


def _permutation_montecarlo_shapley_rollout(
    u: Utility,
    permutation: NDArray[np.int_],
    truncation: TruncationPolicy,
    algorithm_name: str,
    additional_indices: Optional[NDArray[np.int_]] = None,
    use_default_scorer_value: bool = True,
) -> ValuationResult:
    """
    Represents a truncated version of a permutation-based MC estimator. It iterates over
    all subsets starting from the empty set to the full set of indices as specified by
    `permutation`. For each subset, the marginal contribution is computed and added to
    the result. The computation is interrupted if the truncation policy returns `True`.

    !!! Todo
        Reuse in [permutation_montecarlo_shapley()]
        [pydvl.value.shapley.montecarlo.permutation_montecarlo_shapley]

    Args:
        u: Utility object containing model, data, and scoring function.
        permutation: Permutation of indices to be considered.
        truncation: Callable which decides whether to interrupt processing a
            permutation and set all subsequent marginals to zero.
        algorithm_name: For the results object. Used internally by different
            variants of Shapley using this subroutine
        additional_indices: Set of additional indices for data points which should be
            always considered.
        use_default_scorer_value: Use default scorer value even if additional_indices
            is not None.

    Returns:
         ValuationResult object containing computed data values.
    """
    if (
        additional_indices is not None
        and len(np.intersect1d(permutation, additional_indices)) > 0
    ):
        raise ValueError(
            "The class label set and the complement set have to be disjoint."
        )

    result = ValuationResult.zeros(
        algorithm=algorithm_name, indices=u.data.indices, data_names=u.data.data_names
    )

    prev_score = (
        u.default_score
        if (
            use_default_scorer_value
            or additional_indices is None
            or additional_indices is not None
            and len(additional_indices) == 0
        )
        else u(additional_indices)
    )

    truncation_u = u
    if additional_indices is not None:
        # hack to calculate the correct value in reset.
        truncation_indices = np.sort(np.concatenate((permutation, additional_indices)))
        truncation_u = Utility(
            u.model,
            Dataset(
                u.data.x_train[truncation_indices],
                u.data.y_train[truncation_indices],
                u.data.x_test,
                u.data.y_test,
            ),
            u.scorer,
        )
    truncation.reset(truncation_u)

    is_terminated = False
    for i, idx in enumerate(permutation):
        if is_terminated or (is_terminated := truncation(i, prev_score)):
            score = prev_score
        else:
            score = u(
                np.concatenate((permutation[: i + 1], additional_indices))
                if additional_indices is not None and len(additional_indices) > 0
                else permutation[: i + 1]
            )

        marginal = score - prev_score
        result.update(idx, marginal)
        prev_score = score

    return result


def _split_indices_by_label(
    indices: NDArray[np.int_], labels: NDArray[np.int_], label: int
) -> Tuple[NDArray[np.int_], NDArray[np.int_]]:
    """
    Splits the indices into two sets based on the value of `label`, e.g. those samples
    with and without that label.

    Args:
        indices: The indices to be used for referring to the data.
        labels: Corresponding labels for the indices.
        label: Label to be used for splitting.

    Returns:
        Tuple with two sets of indices.
    """
    active_elements = labels == label
    class_indices_set = np.where(active_elements)[0]
    class_complement_indices_set = np.where(~active_elements)[0]
    class_indices_set = indices[class_indices_set]
    class_complement_indices_set = indices[class_complement_indices_set]
    return class_indices_set, class_complement_indices_set



"""
This module implements Group Testing for the approximation of Shapley values, as
introduced in (Jia, R. et al., 2019)[^1]. The sampling of index subsets is
done in such a way that an approximation to the true Shapley values can be
computed with guarantees.

!!! Warning
    This method is very inefficient. Potential improvements to the
    implementation notwithstanding, convergence seems to be very slow (in terms
    of evaluations of the utility required). We recommend other Monte Carlo
    methods instead.

You can read more [in the documentation][computing-data-values].

!!! tip "New in version 0.4.0"

## References

[^1]: <a name="jia_efficient_2019"></a>Jia, R. et al., 2019.
    [Towards Efficient Data Valuation Based on the Shapley
    Value](https://proceedings.mlr.press/v89/jia19a.html).
    In: Proceedings of the 22nd International Conference on Artificial
    Intelligence and Statistics, pp. 1167–1176. PMLR.
"""
import logging
from collections import namedtuple
from typing import Iterable, Optional, Tuple, TypeVar, Union, cast

import cvxpy as cp
import numpy as np
from numpy.random import SeedSequence
from numpy.typing import NDArray
from tqdm.auto import trange

from pydvl.parallel import MapReduceJob, ParallelConfig, effective_n_jobs
from pydvl.utils import Utility
from pydvl.utils.numeric import random_subset_of_size
from pydvl.utils.status import Status
from pydvl.utils.types import Seed, ensure_seed_sequence
from pydvl.value import ValuationResult

__all__ = ["group_testing_shapley", "num_samples_eps_delta"]

log = logging.getLogger(__name__)

T = TypeVar("T", NDArray[np.float_], float)
GTConstants = namedtuple("GTConstants", ["kk", "Z", "q", "q_tot", "T"])


def _constants(
    n: int, epsilon: float, delta: float, utility_range: float
) -> GTConstants:
    """A helper function returning the constants for the algorithm. Pretty ugly,
    yes.

    Args:
        n: The number of data points.
        epsilon: The error tolerance.
        delta: The confidence level.
        utility_range: The range of the utility function.

    Returns:
        A namedtuple with the constants. The fields are the same as in the paper:
            - kk: the sample sizes (i.e. an array of 1, 2, ..., n - 1)
            - Z: the normalization constant
            - q: the probability of drawing a sample of size k
            - q_tot: another normalization constant
            - T: the number of iterations. This will be -1 if the utility_range is
                infinite. E.g. because the [Scorer][pydvl.utils.score.Scorer] does
                not define a range.
    """
    r = utility_range

    kk = np.arange(1, n)  # sample sizes
    Z = 2 * (1.0 / kk).sum()
    q = (1 / kk + 1 / (n - kk)) / Z
    q_tot = (n - 2) / n * q[0] + np.inner(
        q[1:], 1 + 2 * kk[1:] * (kk[1:] - n) / (n * (n - 1))
    )

    def h(u: T) -> T:
        return cast(T, (1 + u) * np.log(1 + u) - u)

    # The implementation in GitHub defines a different bound:
    # T_code = int( 4
    #     / (1 - q_tot**2)
    #     / h(2 * epsilon / Z / r / (1 - q_tot**2))
    #     * np.log(n * (n - 1) / (2 * delta))
    # )
    if r == np.inf:
        log.warning(
            "Group Testing: cannot estimate minimum number of iterations for "
            "unbounded utilities. Please specify a range in the Scorer if "
            "you need this estimate."
        )
        min_iter = -1
    else:
        min_iter = 8 * np.log(n * (n - 1) / (2 * delta)) / (1 - q_tot**2)
        min_iter /= h(2 * epsilon / (np.sqrt(n) * Z * r * (1 - q_tot**2)))

    return GTConstants(kk=kk, Z=Z, q=q, q_tot=q_tot, T=int(min_iter))


def num_samples_eps_delta(
    eps: float, delta: float, n: int, utility_range: float
) -> int:
    r"""Implements the formula in Theorem 3 of (Jia, R. et al., 2019)<sup><a href="#jia_efficient_2019">1</a></sup>
    which gives a lower bound on the number of samples required to obtain an
    (ε/√n,δ/(N(N-1))-approximation to all pair-wise differences of Shapley
    values, wrt. $\ell_2$ norm.

    Args:
        eps: ε
        delta: δ
        n: Number of data points
        utility_range: Range of the [Utility][pydvl.utils.utility.Utility] function
    Returns:
        Number of samples from $2^{[n]}$ guaranteeing ε/√n-correct Shapley
            pair-wise differences of values with probability 1-δ/(N(N-1)).

    !!! tip "New in version 0.4.0"

    """
    constants = _constants(n=n, epsilon=eps, delta=delta, utility_range=utility_range)
    return int(constants.T)


def _group_testing_shapley(
    u: Utility,
    n_samples: int,
    progress: bool = False,
    job_id: int = 1,
    seed: Optional[Union[Seed, SeedSequence]] = None,
):
    """Helper function for
    [group_testing_shapley()][pydvl.value.shapley.gt.group_testing_shapley].

    Computes utilities of sets sampled using the strategy for estimating the
    differences in Shapley values.

    Args:
        u: Utility object with model, data, and scoring function.
        n_samples: total number of samples (subsets) to use.
        progress: Whether to display progress bars for each job.
        job_id: id to use for reporting progress (e.g. to place progres bars)
        seed: Either an instance of a numpy random number generator or a seed for it.
    Returns:

    """
    rng = np.random.default_rng(seed)
    n = len(u.data.indices)
    const = _constants(n, 1, 1, 1)  # don't care about eps,delta,range

    betas: NDArray[np.int_] = np.zeros(
        shape=(n_samples, n), dtype=np.int_
    )  # indicator vars
    uu = np.empty(n_samples)  # utilities

    for t in trange(n_samples, disable=not progress, position=job_id):
        k = rng.choice(const.kk, size=1, p=const.q).item()
        s = random_subset_of_size(u.data.indices, k, seed=rng)
        uu[t] = u(s)
        betas[t, s] = 1
    return uu, betas


def group_testing_shapley(
    u: Utility,
    n_samples: int,
    epsilon: float,
    delta: float,
    *,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
    seed: Optional[Seed] = None,
    **options,
) -> ValuationResult:
    """Implements group testing for approximation of Shapley values as described
    in (Jia, R. et al., 2019)<sup><a href="#jia_efficient_2019">1</a></sup>.

    !!! Warning
        This method is very inefficient. It requires several orders of magnitude
        more evaluations of the utility than others in
        [montecarlo][pydvl.value.shapley.montecarlo]. It also uses several intermediate
        objects like the results from the runners and the constraint matrices
        which can become rather large.

    By picking a specific distribution over subsets, the differences in Shapley
    values can be approximated with a Monte Carlo sum. These are then used to
    solve for the individual values in a feasibility problem.

    Args:
        u: Utility object with model, data, and scoring function
        n_samples: Number of tests to perform. Use
            [num_samples_eps_delta][pydvl.value.shapley.gt.num_samples_eps_delta]
            to estimate this.
        epsilon: From the (ε,δ) sample bound. Use the same as for the
            estimation of `n_iterations`.
        delta: From the (ε,δ) sample bound. Use the same as for the
            estimation of `n_iterations`.
        n_jobs: Number of parallel jobs to use. Each worker performs a chunk
            of all tests (i.e. utility evaluations).
        config: Object configuring parallel computation, with cluster
            address, number of cpus, etc.
        progress: Whether to display progress bars for each job.
        seed: Either an instance of a numpy random number generator or a seed for it.
        options: Additional options to pass to
            [cvxpy.Problem.solve()](https://www.cvxpy.org/tutorial/advanced/index.html#solve-method-options).
            E.g. to change the solver (which defaults to `cvxpy.SCS`) pass
            `solver=cvxpy.CVXOPT`.

    Returns:
        Object with the data values.

    !!! tip "New in version 0.4.0"

    !!! tip "Changed in version 0.5.0"
        Changed the solver to cvxpy instead of scipy's linprog. Added the ability
        to pass arbitrary options to it.
    """

    n = len(u.data.indices)

    const = _constants(
        n=n,
        epsilon=epsilon,
        delta=delta,
        utility_range=u.score_range.max() - u.score_range.min(),
    )
    T = n_samples
    if T < const.T:
        log.warning(
            f"n_samples of {T} are below the required {const.T} for the "
            f"ε={epsilon:.02f} guarantee at δ={1 - delta:.02f} probability"
        )

    samples_per_job = max(1, n_samples // effective_n_jobs(n_jobs, config))

    def reducer(
        results_it: Iterable[Tuple[NDArray, NDArray]]
    ) -> Tuple[NDArray, NDArray]:
        return np.concatenate(list(x[0] for x in results_it)).astype(
            np.float_
        ), np.concatenate(list(x[1] for x in results_it)).astype(np.int_)

    seed_sequence = ensure_seed_sequence(seed)
    map_reduce_seed_sequence, cvxpy_seed = tuple(seed_sequence.spawn(2))

    map_reduce_job: MapReduceJob[Utility, Tuple[NDArray, NDArray]] = MapReduceJob(
        u,
        map_func=_group_testing_shapley,
        reduce_func=reducer,
        map_kwargs=dict(n_samples=samples_per_job, progress=progress),
        config=config,
        n_jobs=n_jobs,
    )
    uu, betas = map_reduce_job(seed=map_reduce_seed_sequence)

    # Matrix of estimated differences. See Eqs. (3) and (4) in the paper.
    C = np.zeros(shape=(n, n))
    for i in range(n):
        for j in range(i + 1, n):
            C[i, j] = np.dot(uu, betas[:, i] - betas[:, j])
    C *= const.Z / T
    total_utility = u(u.data.indices)

    ###########################################################################
    # Solution of the constraint problem with cvxpy

    v = cp.Variable(n)
    constraints = [cp.sum(v) == total_utility]
    for i in range(n):
        for j in range(i + 1, n):
            constraints.append(v[i] - v[j] <= epsilon + C[i, j])
            constraints.append(v[j] - v[i] <= epsilon - C[i, j])

    problem = cp.Problem(cp.Minimize(0), constraints)
    solver = options.pop("solver", cp.SCS)
    problem.solve(solver=solver, **options)

    if problem.status != "optimal":
        log.warning(f"cvxpy returned status {problem.status}")
        values = (
            np.nan * np.ones_like(u.data.indices)
            if not hasattr(v.value, "__len__")
            else v.value
        )
        status = Status.Failed
    else:
        values = v.value
        status = Status.Converged

    return ValuationResult(
        algorithm="group_testing_shapley",
        status=status,
        values=values,
        data_names=u.data.data_names,
        solver_status=problem.status,
    )



"""
This module implements algorithms for the exact and approximate computation of
values and semi-values.

See [Data valuation][computing-data-values] for an introduction to the concepts
and methods implemented here.
"""

from .result import *  # isort: skip
from ..utils import Dataset, Scorer, Utility
from .least_core import *
from .loo import *
from .oob import *
from .sampler import *
from .semivalues import *
from .shapley import *
from .stopping import *



"""
# This module is deprecated

!!! warning "Redirects"
    Imports from this module will be redirected to
    [pydvl.parallel][pydvl.parallel] only until v0.9.0. Please update your
    imports.
"""
import logging

from ..parallel.backend import *
from ..parallel.config import *
from ..parallel.futures import *
from ..parallel.map_reduce import *

log = logging.getLogger(__name__)

# This string for the benefit of deprecation searches:
# remove_in="0.9.0"
log.warning(
    "Importing parallel tools from pydvl.utils is deprecated. "
    "Please import directly from pydvl.parallel. "
    "Redirected imports will be removed in v0.9.0"
)
